# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


import unittest as _ut


class var_integrator_test_case(_ut.TestCase):
    def test_scalar(self):
        from . import (
            make_vars,
            var_ode_sys,
            var_args,
            cos,
            sin,
            par,
            time,
            taylor_adaptive,
            core,
        )
        from sys import getrefcount
        import numpy as np

        x, v = make_vars("x", "v")

        orig_sys = [(x, v), (v, cos(time) - par[0] * v - sin(x))]

        vsys = var_ode_sys(orig_sys, var_args.vars, order=2)

        ta = taylor_adaptive(vsys, [0.2, 0.3], pars=[0.4], time=0.5, compact_mode=True)

        self.assertTrue(ta.dim > 2)
        self.assertEqual(ta.n_orig_sv, 2)
        self.assertTrue(ta.is_variational)
        self.assertEqual(ta.vorder, 2)
        self.assertEqual(ta.vargs, [x, v])

        # Check that the refcount increases by when accessing tstate.
        rc = getrefcount(ta)
        ts = ta.tstate
        self.assertEqual(getrefcount(ta), rc + 1)
        self.assertEqual(ts.shape, (2,))
        self.assertTrue(np.all(ts == [0.0, 0.0]))

        # Test that tstate cannot be written to.
        with self.assertRaises(ValueError) as cm:
            ta.tstate[0] = 0.5

        self.assertEqual(ta.get_vslice(order=0), slice(0, 2, None))
        self.assertEqual(ta.get_vslice(order=0, component=1), slice(1, 2, None))

        self.assertEqual(ta.get_vslice(order=1), slice(2, 6, None))
        self.assertEqual(ta.get_vslice(order=1, component=1), slice(4, 6, None))

        self.assertEqual(ta.get_mindex(0), [0, 0, 0])
        self.assertEqual(ta.get_mindex(1), [1, 0, 0])
        self.assertEqual(ta.get_mindex(2), [0, 1, 0])
        self.assertEqual(ta.get_mindex(3), [0, 0, 1])
        self.assertEqual(ta.get_mindex(4), [1, 1, 0])
        self.assertEqual(ta.get_mindex(i=5), [1, 0, 1])

        ta.propagate_until(3.0)

        ts2 = ta.eval_taylor_map([0.0, 0.0])
        self.assertEqual(getrefcount(ta), rc + 2)
        self.assertTrue(np.shares_memory(ts, ts2))
        self.assertTrue(np.all(ts2 == ta.state[:2]))

        ts2 = ta.eval_taylor_map(np.array([0.0, 0.0]))
        self.assertTrue(np.shares_memory(ts, ts2))
        self.assertTrue(np.all(ts2 == ta.state[:2]))

        with self.assertRaises(TypeError) as cm:
            ta.eval_taylor_map(np.array([0.0, 0.0], dtype=np.int32))
        self.assertTrue(
            "Invalid dtype detected for the inputs of a Taylor map evaluation:"
            in str(cm.exception)
        )

        with self.assertRaises(ValueError) as cm:
            ta.eval_taylor_map(np.array([0.0, 0.0, 0.0, 0.0])[::2])
        self.assertTrue(
            "Invalid inputs array detected in a Taylor map evaluation: the array is not"
            " C-style contiguous, please " in str(cm.exception)
        )

        with self.assertRaises(ValueError) as cm:
            ta.eval_taylor_map(np.array([[0.0, 0.0], [0.0, 0.0]]))
        self.assertTrue(
            "The array of inputs provided for the evaluation of a Taylor map has 2"
            " dimensions, " in str(cm.exception)
        )

        with self.assertRaises(ValueError) as cm:
            ta.eval_taylor_map(np.array([0.0, 0.0, 0.0, 0.0]))
        self.assertTrue(
            "The array of inputs provided for the evaluation of a Taylor map has 4"
            " elements, " in str(cm.exception)
        )

        # Check that passing a view of the integrator's state as input
        # is detected as a memory overlap.
        with self.assertRaises(ValueError) as cm:
            ta.eval_taylor_map(ta.state[:2])
        self.assertTrue("may overlap" in str(cm.exception))

        # Same check with tstate.
        with self.assertRaises(ValueError) as cm:
            ta.eval_taylor_map(ta.tstate)
        self.assertTrue("may overlap" in str(cm.exception))

        if not hasattr(core, "real"):
            return

        from . import real

        prec = 14

        ta = taylor_adaptive(
            vsys,
            [real(0.2, prec), real(0.3, prec)],
            pars=[real(0.4, prec)],
            time=real(0.5, prec),
            compact_mode=True,
            fp_type=real,
        )

        ta.propagate_until(real(3.0, prec))

        with self.assertRaises(ValueError):
            ta.eval_taylor_map(np.empty((2,), dtype=real))

        with self.assertRaises(ValueError):
            ta.eval_taylor_map(np.array([real(0.0, prec), real(0.0, prec - 1)]))

        ts2 = ta.eval_taylor_map(np.array([real(0.0, prec), real(0.0, prec)]))
        self.assertTrue(np.all(ts2 == ta.state[:2]))

    def test_batch(self):
        from . import (
            make_vars,
            var_ode_sys,
            var_args,
            cos,
            sin,
            par,
            time,
            taylor_adaptive_batch,
            core,
        )
        from sys import getrefcount
        import numpy as np

        x, v = make_vars("x", "v")

        orig_sys = [(x, v), (v, cos(time) - par[0] * v - sin(x))]

        vsys = var_ode_sys(orig_sys, var_args.vars, order=2)

        ta = taylor_adaptive_batch(
            vsys,
            [[0.2, 0.21], [0.3, 0.31]],
            pars=[[0.4, 0.41]],
            time=[0.5, 0.51],
            compact_mode=True,
        )

        self.assertTrue(ta.dim > 2)
        self.assertEqual(ta.n_orig_sv, 2)
        self.assertTrue(ta.is_variational)
        self.assertEqual(ta.vorder, 2)
        self.assertEqual(ta.vargs, [x, v])

        # Check that the refcount increases by when accessing tstate.
        rc = getrefcount(ta)
        ts = ta.tstate
        self.assertEqual(getrefcount(ta), rc + 1)
        self.assertEqual(ts.shape, (2, 2))
        self.assertTrue(np.all(ts == [[0.0, 0.0], [0.0, 0.0]]))

        # Test that tstate cannot be written to.
        with self.assertRaises(ValueError) as cm:
            ta.tstate[0] = 0.5

        self.assertEqual(ta.get_vslice(order=0), slice(0, 2, None))
        self.assertEqual(ta.get_vslice(order=0, component=1), slice(1, 2, None))

        self.assertEqual(ta.get_vslice(order=1), slice(2, 6, None))
        self.assertEqual(ta.get_vslice(order=1, component=1), slice(4, 6, None))

        self.assertEqual(ta.get_mindex(0), [0, 0, 0])
        self.assertEqual(ta.get_mindex(1), [1, 0, 0])
        self.assertEqual(ta.get_mindex(2), [0, 1, 0])
        self.assertEqual(ta.get_mindex(3), [0, 0, 1])
        self.assertEqual(ta.get_mindex(4), [1, 1, 0])
        self.assertEqual(ta.get_mindex(i=5), [1, 0, 1])

        ta.propagate_until(3.0)

        ts2 = ta.eval_taylor_map([[0.0, 0.0], [0.0, 0.0]])
        self.assertEqual(getrefcount(ta), rc + 2)
        self.assertTrue(np.shares_memory(ts, ts2))
        self.assertTrue(np.all(ts2 == ta.state[:2]))

        ts2 = ta.eval_taylor_map(np.array([[0.0, 0.0], [0.0, 0.0]]))
        self.assertTrue(np.shares_memory(ts, ts2))
        self.assertTrue(np.all(ts2 == ta.state[:2]))

        with self.assertRaises(TypeError) as cm:
            ta.eval_taylor_map(np.array([0.0, 0.0], dtype=np.int32))
        self.assertTrue(
            "Invalid dtype detected for the inputs of a Taylor map evaluation:"
            in str(cm.exception)
        )

        with self.assertRaises(ValueError) as cm:
            ta.eval_taylor_map(np.array([0.0, 0.0, 0.0, 0.0])[::2])
        self.assertTrue(
            "Invalid inputs array detected in a Taylor map evaluation: the array is not"
            " C-style contiguous, please " in str(cm.exception)
        )

        with self.assertRaises(ValueError) as cm:
            ta.eval_taylor_map(np.array([0.0, 0.0]))
        self.assertTrue(
            "The array of inputs provided for the evaluation of a Taylor map has 1"
            " dimension(s), " in str(cm.exception)
        )

        with self.assertRaises(ValueError) as cm:
            ta.eval_taylor_map(np.array([[0.0, 0.0]]))
        self.assertTrue(
            "The array of inputs provided for the evaluation of a Taylor map has 1"
            " row(s), but it must have 2 row(s) instead" in str(cm.exception)
        )

        with self.assertRaises(ValueError) as cm:
            ta.eval_taylor_map(np.array([[0.0], [0.0]]))
        self.assertTrue(
            "The array of inputs provided for the evaluation of a Taylor map has 1"
            " column(s), but it must have 2 column(s) instead" in str(cm.exception)
        )

        # Check that passing a view of the integrator's state as input
        # is detected as a memory overlap.
        with self.assertRaises(ValueError) as cm:
            ta.eval_taylor_map(ta.state[:2])
        self.assertTrue("may overlap" in str(cm.exception))

        # Same check with tstate.
        with self.assertRaises(ValueError) as cm:
            ta.eval_taylor_map(ta.tstate)
        self.assertTrue("may overlap" in str(cm.exception))

    def test_size_check_bug(self):
        # BUG: wrong size check on the input for a Taylor map (the size is checked against
        # the number of original state variables instead of the number of variational arguments).

        from . import (
            make_vars,
            var_ode_sys,
            var_args,
            cos,
            sin,
            par,
            time,
            taylor_adaptive_batch,
            taylor_adaptive,
            core,
        )
        import numpy as np

        x, v = make_vars("x", "v")

        orig_sys = [(x, v), (v, cos(time) - par[0] * v - sin(x))]

        vsys = var_ode_sys(orig_sys, var_args.vars | var_args.params, order=2)

        ta = taylor_adaptive(
            vsys,
            [0.2, 0.3],
            pars=[0.4],
            time=0.5,
            compact_mode=True,
        )

        # This would throw.
        self.assertTrue(np.all(ta.state[:2] == ta.eval_taylor_map([0.0, 0.0, 0.0])))

        # Test the batch case too.
        ta = taylor_adaptive_batch(
            vsys,
            [[0.2, 0.21], [0.3, 0.31]],
            pars=[[0.4, 0.41]],
            time=[0.5, 0.51],
            compact_mode=True,
        )

        # This would throw.
        self.assertTrue(
            np.all(
                ta.state[:2] == ta.eval_taylor_map([[0.0, 0.0], [0.0, 0.0], [0.0, 0.0]])
            )
        )



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest as _ut


class mp_test_case(_ut.TestCase):
    def test_cfunc(self):
        from . import core

        if not hasattr(core, "real"):
            return

        from . import real, cfunc, make_vars, sin, par, time
        import numpy as np

        x, y = make_vars("x", "y")
        func = [sin(x + y), x - par[0], x + y + par[1] + time]

        with self.assertRaises(ValueError) as cm:
            cfunc(func, vars=[y, x], fp_type=real, batch_size=2)
        self.assertTrue(
            "Batch sizes greater than 1 are not supported for this floating-point type"
            in str(cm.exception)
        )

        with self.assertRaises(ValueError) as cm:
            cfunc(func, vars=[y, x], fp_type=real, prec=-1)
        self.assertTrue(
            "An invalid precision value of -1 was passed to make_multi_cfunc()"
            in str(cm.exception)
        )

        prec = 237

        fn = cfunc(func, vars=[y, x], fp_type=real, prec=prec)

        # Initial simple test.
        inputs = np.array([real(1, prec), real(2, prec)])
        pars = np.array([real(3, prec), real(4, prec)])
        out = fn(inputs=inputs, pars=pars, time=real("1.1", prec))
        self.assertEqual(out[0], np.sin(inputs[1] + inputs[0]))
        self.assertEqual(out[1], inputs[1] - pars[0])
        self.assertEqual(out[2], inputs[1] + inputs[0] + pars[1] + real("1.1", prec))

        # With provided empty output.
        out = fn(
            inputs=inputs,
            pars=pars,
            outputs=np.empty((3,), dtype=real),
            time=real("1.1", prec),
        )
        self.assertEqual(out[0], np.sin(inputs[1] + inputs[0]))
        self.assertEqual(out[1], inputs[1] - pars[0])
        self.assertEqual(out[2], inputs[1] + inputs[0] + pars[1] + real("1.1", prec))

        # Test with non-owning arrays.
        out = fn(inputs=inputs[:], pars=pars[:], time=real("1.1", prec))
        self.assertEqual(out[0], np.sin(inputs[1] + inputs[0]))
        self.assertEqual(out[1], inputs[1] - pars[0])
        self.assertEqual(out[2], inputs[1] + inputs[0] + pars[1] + real("1.1", prec))

        # Test with overlapping arrays.
        with self.assertRaises(ValueError) as cm:
            fn(inputs=inputs, pars=inputs, time=real("1.1", prec))
        self.assertTrue(
            "Potential memory overlaps detected when attempting to evaluate a compiled"
            " function: please make sure that all input arrays are distinct"
            in str(cm.exception)
        )

        # Test with non-contiguous arrays.
        inputs = np.array([real(1, prec), 0, real(2, prec), 0])
        pars = np.array([real(3, prec), 0, real(4, prec), 0])
        inputs = inputs[::2]
        pars = pars[::2]
        with self.assertRaises(ValueError) as cm:
            fn(inputs=inputs, pars=pars, time=real("1.1", prec))
        self.assertTrue(
            "Invalid inputs array detected: the array is not C-style contiguous, please"
            " consider using numpy.ascontiguousarray() to turn it into one"
            in str(cm.exception)
        )
        with self.assertRaises(ValueError) as cm:
            fn(inputs=np.ascontiguousarray(inputs), pars=pars, time=real("1.1", prec))
        self.assertTrue(
            "Invalid parameters array detected: the array is not C-style contiguous,"
            " please consider using numpy.ascontiguousarray() to turn it into one"
            in str(cm.exception)
        )

        # Test in single eval mode with wrong precision for the time value.
        inputs = np.array([real(1, prec), real(2, prec)])
        pars = np.array([real(3, prec), real(4, prec)])
        with self.assertRaises(ValueError) as cm:
            fn(inputs=inputs, pars=pars, time=real("1.1", prec - 1))
        self.assertTrue(
            "An invalid time value was passed for the evaluation of a compiled"
            " function in multiprecision mode: the time value has a precision of"
            f" {prec-1}, while the expected precision is {prec} instead"
            in str(cm.exception)
        )

        # Test multieval too.
        inputs = np.array(
            [[real(1, prec), real(-1, prec)], [real(2, prec), real(-2, prec)]]
        )
        pars = np.array(
            [[real(3, prec), real(-3, prec)], [real(4, prec), real(-4, prec)]]
        )
        tm = np.array([real("1.1", prec), real("1.3", prec)])
        out = fn(inputs=inputs, pars=pars, time=tm)
        self.assertEqual(out[0, 0], np.sin(inputs[1, 0] + inputs[0, 0]))
        self.assertEqual(out[1, 0], inputs[1, 0] - pars[0, 0])
        self.assertEqual(out[2, 0], inputs[1, 0] + inputs[0, 0] + pars[1, 0] + tm[0])
        self.assertEqual(out[0, 1], np.sin(inputs[1, 1] + inputs[0, 1]))
        self.assertEqual(out[1, 1], inputs[1, 1] - pars[0, 1])
        self.assertEqual(out[2, 1], inputs[1, 1] + inputs[0, 1] + pars[1, 1] + tm[1])

        # With provided empty output.
        out = fn(
            inputs=inputs, pars=pars, outputs=np.empty((3, 2), dtype=real), time=tm
        )
        self.assertEqual(out[0, 0], np.sin(inputs[1, 0] + inputs[0, 0]))
        self.assertEqual(out[1, 0], inputs[1, 0] - pars[0, 0])
        self.assertEqual(out[2, 0], inputs[1, 0] + inputs[0, 0] + pars[1, 0] + tm[0])
        self.assertEqual(out[0, 1], np.sin(inputs[1, 1] + inputs[0, 1]))
        self.assertEqual(out[1, 1], inputs[1, 1] - pars[0, 1])
        self.assertEqual(out[2, 1], inputs[1, 1] + inputs[0, 1] + pars[1, 1] + tm[1])

        # Non-owning.
        out = fn(inputs=inputs[:], pars=pars[:], time=tm[:])
        self.assertEqual(out[0, 0], np.sin(inputs[1, 0] + inputs[0, 0]))
        self.assertEqual(out[1, 0], inputs[1, 0] - pars[0, 0])
        self.assertEqual(out[2, 0], inputs[1, 0] + inputs[0, 0] + pars[1, 0] + tm[0])
        self.assertEqual(out[0, 1], np.sin(inputs[1, 1] + inputs[0, 1]))
        self.assertEqual(out[1, 1], inputs[1, 1] - pars[0, 1])
        self.assertEqual(out[2, 1], inputs[1, 1] + inputs[0, 1] + pars[1, 1] + tm[1])

        # Non-contiguous.
        inputs = np.array(
            [
                [real(1, prec), 0, real(-1, prec), 0],
                [real(2, prec), 0, real(-2, prec), 0],
            ]
        )
        pars = np.array(
            [
                [real(3, prec), 0, real(-3, prec), 0],
                [real(4, prec), 0, real(-4, prec), 0],
            ]
        )
        tm = np.array([real("1.1", prec), 0, real("1.3", prec), 0])

        inputs = inputs[:, ::2]
        pars = pars[:, ::2]
        tm = tm[::2]
        with self.assertRaises(ValueError) as cm:
            fn(inputs=inputs[:], pars=pars[:], time=tm)
        self.assertTrue(
            "Invalid inputs array detected: the array is not C-style contiguous, please"
            " consider using numpy.ascontiguousarray() to turn it into one"
            in str(cm.exception)
        )
        with self.assertRaises(ValueError) as cm:
            fn(inputs=np.ascontiguousarray(inputs[:]), pars=pars[:], time=tm)
        self.assertTrue(
            "Invalid parameters array detected: the array is not C-style contiguous,"
            " please consider using numpy.ascontiguousarray() to turn it into one"
            in str(cm.exception)
        )

        # Error modes.
        inputs = np.array([real(1, prec), real(2, prec)])
        pars = np.array([real(3, prec), real(4, prec - 1)])
        with self.assertRaises(ValueError) as cm:
            fn(inputs=inputs, pars=pars, time=real(42, prec))
        self.assertTrue(
            "A real with precision 236 was detected at the indices" in str(cm.exception)
        )

        inputs = np.array([real(1, prec), real(2, prec)])
        pars = np.array([real(3, prec), real(4, prec)])
        with self.assertRaises(ValueError) as cm:
            fn(inputs=inputs, pars=pars, time=real(42, prec - 2))
        self.assertTrue(
            "An invalid time value was passed for the evaluation of a compiled"
            " function in multiprecision mode: the time value has a precision of"
            f" {prec - 2}, while the expected precision is {prec} instead"
            in str(cm.exception)
        )

        pars = np.array([real(3, prec), real(4, prec)])
        with self.assertRaises(ValueError) as cm:
            fn(inputs=[1, 2], pars=pars, time=real(42, prec))
        self.assertTrue(
            "in an array which should instead contain elements with a precision of 237"
            in str(cm.exception)
        )

    def test_sympy(self):
        from . import core

        if not hasattr(core, "real"):
            return

        try:
            import sympy
            from mpmath import mp, pi, workprec
        except ImportError:
            return

        from . import to_sympy, from_sympy, real, expression, sum as sum_hy

        with workprec(128):
            from_spy_ex = from_sympy(pi() + sympy.Symbol("x"))
            self.assertEqual(
                sum_hy(
                    [
                        expression(
                            real("3.141592653589793238462643383279502884195", 128)
                        ),
                        expression("x"),
                    ]
                ),
                from_spy_ex,
            )

            self.assertEqual(
                to_sympy(from_sympy(pi() + sympy.Symbol("x"))),
                pi() + sympy.Symbol("x", real=True),
            )

        self.assertEqual(
            str(from_sympy(sympy.Integer(2**500 + 1))),
            "3.2733906078961418700131896968275991522166420460430647894832913680961337964046745548832700923259041571508866841275600710092172565458853930533285275893770e+150",
        )

        self.assertEqual(
            str(
                from_sympy(
                    sympy.Rational(sympy.Integer(2**128 + 1), sympy.Integer(2**500))
                )
            ),
            "1.039540976564489921853040458068190636782e-112",
        )

        self.assertEqual(
            from_sympy(sympy.Integer((2 << 128) + 1)),
            expression(real("680564733841876926926749214863536422913.0", 130)),
        )

    def test_expression(self):
        from . import core

        if not hasattr(core, "real"):
            return

        from . import expression as ex, real, kepE, kepF, kepDE, atan2

        self.assertEqual(
            str(ex(real("1.1", 128))), "1.100000000000000000000000000000000000001"
        )

        self.assertEqual(
            str(1.0 + ex(real("1.1", 128))), "2.099999999999999999999999999999999999995"
        )
        self.assertEqual(
            str(ex(real("1.1", 128)) + 1), "2.099999999999999999999999999999999999995"
        )

        self.assertEqual(
            str(1.0 - ex(real("1.1", 128))),
            "-1.000000000000000000000000000000000000012e-1",
        )
        self.assertEqual(
            str(ex(real("1.1", 128)) - 1),
            "1.000000000000000000000000000000000000012e-1",
        )

        self.assertEqual(
            str(1.0 * ex(real("1.1", 128))),
            "1.100000000000000000000000000000000000001",
        )
        self.assertEqual(
            str(ex(real("1.1", 128)) * 1),
            "1.100000000000000000000000000000000000001",
        )

        self.assertEqual(
            str(1.0 / ex(real("1.1", 128))),
            "9.090909090909090909090909090909090909070e-1",
        )
        self.assertEqual(
            str(ex(real("1.1", 128)) / 1),
            "1.100000000000000000000000000000000000001",
        )

        # NOTE: just test execution for the time being,
        # as these currently do not fold.
        ex(1.1) ** real("1.1", 128)

        kepE(ex("x"), real("1.1", 128))
        kepE(real("1.1", 128), ex("x"))

        kepF(ex("x"), real("1.1", 128), ex("y"))
        kepF(real("1.1", 128), ex("y"), ex("x"))

        kepDE(ex("x"), real("1.1", 128), ex("y"))
        kepDE(real("1.1", 128), ex("y"), ex("x"))

        atan2(ex("x"), real("1.1", 128))
        atan2(real("1.1", 128), ex("x"))

    def test_events(self):
        from . import core

        if not hasattr(core, "real"):
            return

        # Basic event testing.
        from . import make_vars, taylor_adaptive, real, t_event

        x, v = make_vars("x", "v")

        prec = 237

        ta = taylor_adaptive(
            [(x, v), (v, -x)],
            [real(0, prec), real(1, prec)],
            fp_type=real,
            t_events=[t_event(v, fp_type=real)],
        )

        ta.propagate_until(real(100))

        self.assertLess(
            abs(
                ta.time
                - real(
                    "1.570796326794896619231321691639751442098584699687552910487472296153908199",
                    prec,
                )
            ),
            1e-70,
        )
        self.assertLess(abs(ta.state[0] - 1), 1e-70)
        self.assertLess(abs(ta.state[1]), 1e-70)

    def test_c_out(self):
        from . import core

        if not hasattr(core, "real"):
            return

        from . import make_vars, taylor_adaptive, real, core
        import numpy as np
        from copy import deepcopy

        x, v = make_vars("x", "v")

        prec = 237

        ta = taylor_adaptive(
            [(x, v), (v, -x)],
            [real(0, prec), real(1, prec)],
            fp_type=real,
        )

        res = ta.propagate_until(real(10), c_output=True)
        self.assertTrue(res[-1] is None)
        r5 = res[-2](real("5.3", prec + 10))
        self.assertFalse(r5.flags.writeable)
        self.assertFalse(r5.flags.owndata)
        x_cmp = np.sin(real("5.3", prec))
        v_cmp = np.cos(real("5.3", prec))
        self.assertLess(abs((x_cmp - r5[0]) / x_cmp), 1e-70)
        self.assertLess(abs((v_cmp - r5[1]) / v_cmp), 1e-70)

        # Test for the vector time overload.
        r5 = deepcopy(r5)
        r7 = deepcopy(res[-2](real("7.3", prec + 10)))
        vout = res[-2]([real("5.3", prec + 10), real("7.3", prec + 10)])
        self.assertTrue(np.all(vout[0] == r5))
        self.assertTrue(np.all(vout[1] == r7))

        # Test with non-owning array.
        vout = res[-2](
            core._make_no_real_array([real("5.3", prec + 10), real("7.3", prec + 10)])
        )
        self.assertTrue(np.all(vout[0] == r5))
        self.assertTrue(np.all(vout[1] == r7))

        # Test failure modes.
        arr = np.empty((2,), dtype=real)
        with self.assertRaises(ValueError) as cm:
            res[-2](arr)
        self.assertTrue("A non-constructed/invalid real" in str(cm.exception))
        self.assertTrue("0" in str(cm.exception))

        arr[0] = 1
        with self.assertRaises(ValueError) as cm:
            res[-2](arr)
        self.assertTrue("A non-constructed/invalid real" in str(cm.exception))
        self.assertTrue("1" in str(cm.exception))

    def test_basic(self):
        from . import core

        if not hasattr(core, "real"):
            return

        from . import make_vars, taylor_adaptive, sin, real, core
        import numpy as np

        x, v = make_vars("x", "v")

        prec = 237

        # Small test with automatic inference of precision.
        ta = taylor_adaptive(
            [(x, v), (v, -9.8 * sin(x))],
            [real(-1, prec), real(0, prec)],
            fp_type=real,
        )

        self.assertTrue(ta.compact_mode)

        self.assertEqual(ta.prec, prec)

        def compute_energy(sv):
            return (sv[1] * sv[1]) / 2 + 9.8 * (1 - np.cos(sv[0]))

        orig_E = compute_energy(ta.state)

        ta.propagate_until(real(10))

        self.assertLess(abs((orig_E - compute_energy(ta.state)) / orig_E), 1e-71)

        orig_sv = ta.state

        # Precision explicitly given.
        ta = taylor_adaptive(
            [(x, v), (v, -9.8 * sin(x))],
            [real(-1, prec - 1), real(0, prec + 1)],
            fp_type=real,
            prec=prec,
        )

        self.assertEqual(ta.prec, prec)

        # Compact mode disabled.
        ta2 = taylor_adaptive(
            [(x, v), (v, -9.8 * sin(x))],
            [real(-1, prec - 1), real(0, prec + 1)],
            fp_type=real,
            prec=10,
            compact_mode=False,
        )

        self.assertFalse(ta2.compact_mode)

        ta.propagate_until(real(10))
        self.assertTrue(np.all(orig_sv == ta.state))

        # Test with bogus precision.
        with self.assertRaises(ValueError) as cm:
            taylor_adaptive(
                [(x, v), (v, -9.8 * sin(x))],
                [real(-1, prec - 1), real(0, prec + 1)],
                fp_type=real,
                prec=-1,
                compact_mode=False,
            )
        self.assertTrue(
            "Cannot set the precision of a real to the value -1" in str(cm.exception)
        )

        # Testing for the pyreal_check_array() helper.
        arr = np.empty((5,), dtype=real)
        with self.assertRaises(ValueError) as cm:
            core._real_check_array(arr)
        self.assertTrue("A non-constructed/invalid real" in str(cm.exception))
        self.assertTrue("0" in str(cm.exception))

        arr[0] = 5
        with self.assertRaises(ValueError) as cm:
            core._real_check_array(arr)
        self.assertTrue("A non-constructed/invalid real" in str(cm.exception))
        self.assertTrue("1" in str(cm.exception))

        core._real_check_array(core._make_no_real_array(arr))

        arr.fill(real(5))
        core._real_check_array(arr)

        arr = np.empty((5, 5), dtype=real)
        with self.assertRaises(ValueError) as cm:
            core._real_check_array(arr)
        self.assertTrue("A non-constructed/invalid real" in str(cm.exception))
        self.assertTrue("0, 0" in str(cm.exception))

        arr[0, 0] = 5
        with self.assertRaises(ValueError) as cm:
            core._real_check_array(arr)
        self.assertTrue("A non-constructed/invalid real" in str(cm.exception))
        self.assertTrue("0, 1" in str(cm.exception))
        arr[0].fill(real(5))
        arr[1, 0] = 6
        arr[1, 1] = 7
        with self.assertRaises(ValueError) as cm:
            core._real_check_array(arr)
        self.assertTrue("A non-constructed/invalid real" in str(cm.exception))
        self.assertTrue("1, 2" in str(cm.exception))

        arr.fill(real(5))
        core._real_check_array(arr)

        arr = np.empty((), dtype=real)
        core._real_check_array(arr)

        arr = np.empty((1, 0, 3), dtype=real)
        core._real_check_array(arr)

        arr = np.empty((1, 2, 3), dtype=real)

        with self.assertRaises(ValueError) as cm:
            core._real_check_array(arr)
        self.assertTrue(
            "Cannot call pyreal_check_array() on an array with 3 dimensions"
            in str(cm.exception)
        )

        # Testing for the pyreal_ensure_array() helper.
        arr = np.empty((5,), dtype=real)
        core._real_ensure_array(arr, 71)
        for val in arr:
            self.assertEqual(val, 0)
            self.assertEqual(val.prec, 71)

        arr = np.empty((5,), dtype=real)
        arr[0] = real(5, 88)
        core._real_ensure_array(arr, 71)
        self.assertEqual(arr[0], 5)
        self.assertEqual(arr[0].prec, 71)
        for val in arr[1:]:
            self.assertEqual(val, 0)
            self.assertEqual(val.prec, 71)

        arr = core._make_no_real_array(np.empty((5,), dtype=real))
        arr[0] = real(5, 88)
        core._real_ensure_array(arr, 71)
        self.assertEqual(arr[0], 5)
        self.assertEqual(arr[0].prec, 71)
        for val in arr[1:]:
            self.assertEqual(val, 0)
            self.assertEqual(val.prec, 71)

        arr = np.full((5,), real("1.1", 11))
        core._real_ensure_array(arr, 71)
        for val in arr:
            self.assertEqual(val, real(real("1.1", 11), 71))
            self.assertEqual(val.prec, 71)

        arr = np.empty((5, 5), dtype=real)
        core._real_ensure_array(arr, 71)
        for r in arr:
            for val in r:
                self.assertEqual(val, 0)
                self.assertEqual(val.prec, 71)

        arr = np.empty((5, 5), dtype=real)
        arr[0, :] = real(5, 88)
        core._real_ensure_array(arr, 71)
        self.assertTrue(np.all(arr[0, :] == np.full((5,), real(5, 88))))
        for r in arr[1:]:
            for val in r:
                self.assertEqual(val, 0)
                self.assertEqual(val.prec, 71)

        arr = np.full((5, 5), real("1.1", 11))
        core._real_ensure_array(arr, 71)
        for r in arr:
            for val in r:
                self.assertEqual(val, real(real("1.1", 11), 71))
                self.assertEqual(val.prec, 71)

        arr = np.empty((), dtype=real)
        core._real_ensure_array(arr, 71)

        arr = np.empty((1, 0, 3), dtype=real)
        self.assertEqual(arr.size, 0)

        # Empty init state with explicit precision.
        ta = taylor_adaptive(
            [(x, v), (v, -9.8 * sin(x))],
            fp_type=real,
            prec=prec,
        )

        self.assertEqual(ta.prec, prec)
        self.assertTrue(np.all(ta.state == [0, 0]))

        ta = taylor_adaptive(
            [(x, v), (v, -9.8 * sin(x))],
            [],
            fp_type=real,
            prec=prec,
        )

        self.assertEqual(ta.prec, prec)
        self.assertTrue(np.all(ta.state == [0, 0]))

        # Empty init state without explicit precision.
        with self.assertRaises(ValueError) as cm:
            ta = taylor_adaptive(
                [(x, v), (v, -9.8 * sin(x))],
                fp_type=real,
            )
        self.assertTrue("we cannot deduce the desired precision" in str(cm.exception))



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Version setup.
from ._version import __version__

import cloudpickle as _cloudpickle
from threading import Lock as _Lock

# We import the sub-modules into the root namespace.
from .core import *

# Explicitly import the submodules
# NOTE: it is *important* that the import is performed
# here, *after* the initial import of core. Otherwise,
# we would get missing symbols on POSIX platforms.
from . import test, model, callback


def _with_real128():
    # Small helper to check if real128 is available.
    from . import core

    return hasattr(core, "real128")


def _with_real():
    # Small helper to check if real is available.
    from . import core

    return hasattr(core, "real")


from numpy import float32 as _f32, float64 as _f64, longdouble as _ld, dtype as _dtype

_fp_to_suffix_dict = {_f32: "_flt", _f64: "_dbl", float: "_dbl", _ld: "_ldbl"}

del _f32
del _f64
del _ld

if _with_real128():
    _fp_to_suffix_dict[real128] = "_f128"

if _with_real():
    _fp_to_suffix_dict[real] = "_real"


def _fp_to_suffix(fp_t):
    if not isinstance(fp_t, type):
        raise TypeError(
            'A Python type was expected in input, but an object of type "{}" was'
            " provided instead".format(type(fp_t))
        )

    if fp_t in _fp_to_suffix_dict:
        return _fp_to_suffix_dict[fp_t]

    raise TypeError(
        'The floating-point type "{}" is not recognized/supported'.format(fp_t)
    )


def taylor_adaptive(sys, state=[], **kwargs):
    from . import core

    fp_type = kwargs.pop("fp_type", float)
    fp_suffix = _fp_to_suffix(fp_type)

    return getattr(core, "taylor_adaptive{}".format(fp_suffix))(sys, state, **kwargs)


def taylor_adaptive_batch(sys, state, **kwargs):
    from . import core

    fp_type = kwargs.pop("fp_type", float)
    fp_suffix = _fp_to_suffix(fp_type)

    return getattr(core, "taylor_adaptive_batch{}".format(fp_suffix))(
        sys, state, **kwargs
    )


def recommended_simd_size(fp_type=float):
    from . import core

    fp_suffix = _fp_to_suffix(fp_type)

    return getattr(core, "_recommended_simd_size{}".format(fp_suffix))()


def cfunc(fn, vars, **kwargs):
    from . import core

    fp_type = kwargs.pop("fp_type", float)
    fp_suffix = _fp_to_suffix(fp_type)

    return getattr(core, "cfunc{}".format(fp_suffix))(fn, vars, **kwargs)


def nt_event(ex, callback, **kwargs):
    from . import core

    fp_type = kwargs.pop("fp_type", float)
    fp_suffix = _fp_to_suffix(fp_type)

    return getattr(core, "nt_event{}".format(fp_suffix))(ex, callback, **kwargs)


def t_event(ex, **kwargs):
    from . import core

    fp_type = kwargs.pop("fp_type", float)
    fp_suffix = _fp_to_suffix(fp_type)

    return getattr(core, "t_event{}".format(fp_suffix))(ex, **kwargs)


def nt_event_batch(ex, callback, **kwargs):
    from . import core

    fp_type = kwargs.pop("fp_type", float)
    fp_suffix = _fp_to_suffix(fp_type)

    return getattr(core, "nt_event_batch{}".format(fp_suffix))(ex, callback, **kwargs)


def t_event_batch(ex, **kwargs):
    from . import core

    fp_type = kwargs.pop("fp_type", float)
    fp_suffix = _fp_to_suffix(fp_type)

    return getattr(core, "t_event_batch{}".format(fp_suffix))(ex, **kwargs)


def from_sympy(ex, s_dict={}):
    from ._sympy_utils import _with_sympy, _from_sympy_impl

    if not _with_sympy:
        raise ImportError(
            "The 'from_sympy()' function is not available because sympy is not"
            " installed"
        )

    from sympy import Basic
    from .core import expression

    if not isinstance(ex, Basic):
        raise TypeError(
            "The 'ex' parameter must be a sympy expression but it is of type {} instead".format(
                type(ex)
            )
        )

    if not isinstance(s_dict, dict):
        raise TypeError(
            "The 's_dict' parameter must be a dict but it is of type {} instead".format(
                type(s_dict)
            )
        )

    if any(not isinstance(_, Basic) for _ in s_dict):
        raise TypeError("The keys in 's_dict' must all be sympy expressions")

    if any(not isinstance(s_dict[_], expression) for _ in s_dict):
        raise TypeError("The values in 's_dict' must all be heyoka expressions")

    return _from_sympy_impl(ex, s_dict, {})


# Machinery for the setup of the serialization backend.


# Helper to create dicts mapping a name to a serialization backend
# and vice-versa.
def _make_s11n_backend_maps():
    import pickle

    ret = {"cloudpickle": _cloudpickle, "pickle": pickle}

    try:
        import dill

        ret["dill"] = dill
    except ImportError:
        pass

    inv = dict([(ret[_], _) for _ in ret])

    return ret, inv


_s11n_backend_map, _s11n_backend_inv_map = _make_s11n_backend_maps()

# The currently active s11n backend.
_s11n_backend = _cloudpickle

# Lock to protect access to _s11n_backend.
_s11n_backend_mutex = _Lock()


def set_serialization_backend(name):
    global _s11n_backend

    if not isinstance(name, str):
        raise TypeError(
            "The serialization backend must be specified as a string, but an object of"
            " type {} was provided instead".format(type(name))
        )

    if not name in _s11n_backend_map:
        raise ValueError(
            "The serialization backend '{}' is not valid. The valid backends are: {}".format(
                name, list(_s11n_backend_map.keys())
            )
        )

    new_backend = _s11n_backend_map[name]

    with _s11n_backend_mutex:
        _s11n_backend = new_backend


def get_serialization_backend():
    with _s11n_backend_mutex:
        return _s11n_backend


# Machinery to setup the custom SSL verify file.
def _setup_custom_verify_file():
    try:
        import certifi
    except ImportError:
        return

    from .core import _set_ssl_verify_file

    _set_ssl_verify_file(certifi.where())


_setup_custom_verify_file()


# Ensemble propagations.
def _ensemble_propagate_generic(tp, ta, arg, n_iter, gen, **kwargs):
    import numpy as np

    if not isinstance(n_iter, int):
        raise TypeError(
            "The n_iter parameter must be an integer, but an object of type {} was"
            " provided instead".format(type(n_iter))
        )

    if n_iter < 0:
        raise ValueError(
            "The n_iter parameter must be non-negative, but it is {} instead".format(
                n_iter
            )
        )

    # Validate arg and max_delta_t, if present.
    def is_iterable(x):
        from collections.abc import Iterable

        return isinstance(x, Iterable)

    if tp == "until" or tp == "for":
        if is_iterable(arg):
            raise TypeError(
                "Cannot perform an ensemble propagate_until/for(): the final epoch/time"
                " interval must be a scalar, not an iterable object"
            )
    else:
        arg = np.array(arg)

        if arg.ndim != 1:
            raise ValueError(
                "Cannot perform an ensemble propagate_grid(): the input time grid must"
                " be one-dimensional, but instead it has {} dimensions".format(arg.ndim)
            )

    if "max_delta_t" in kwargs and is_iterable(kwargs["max_delta_t"]):
        raise TypeError(
            'Cannot perform an ensemble propagate_until/for/grid(): the "max_delta_t"'
            " argument must be a scalar, not an iterable object"
        )

    # Parallelisation algorithm.
    algo = kwargs.pop("algorithm", "thread")
    allowed_algos = ["thread", "process"]

    if algo == "thread":
        from ._ensemble_impl import _ensemble_propagate_thread

        return _ensemble_propagate_thread(tp, ta, arg, n_iter, gen, **kwargs)

    if algo == "process":
        from ._ensemble_impl import _ensemble_propagate_process

        return _ensemble_propagate_process(tp, ta, arg, n_iter, gen, **kwargs)

    raise ValueError(
        "The parallelisation algorithm must be one of {}, but '{}' was provided instead".format(
            allowed_algos, algo
        )
    )


def ensemble_propagate_until(ta, t, n_iter, gen, **kwargs):
    return _ensemble_propagate_generic("until", ta, t, n_iter, gen, **kwargs)


def ensemble_propagate_for(ta, delta_t, n_iter, gen, **kwargs):
    return _ensemble_propagate_generic("for", ta, delta_t, n_iter, gen, **kwargs)


def ensemble_propagate_grid(ta, grid, n_iter, gen, **kwargs):
    return _ensemble_propagate_generic("grid", ta, grid, n_iter, gen, **kwargs)


def ensemble_propagate_until_batch(ta, t, n_iter, gen, **kwargs):
    return _ensemble_propagate_generic("until", ta, t, n_iter, gen, **kwargs)


def ensemble_propagate_for_batch(ta, delta_t, n_iter, gen, **kwargs):
    return _ensemble_propagate_generic("for", ta, delta_t, n_iter, gen, **kwargs)


def ensemble_propagate_grid_batch(ta, grid, n_iter, gen, **kwargs):
    return _ensemble_propagate_generic("grid", ta, grid, n_iter, gen, **kwargs)


def _real_reduce_factory():
    # Internal factory function used in the implementation
    # of the pickle protocol for real.
    return real()


# Machinery for the par generator.
def _create_par():
    from . import core

    return core._par_generator()


par = _create_par()
"""
Parameter factory.

This global object is used to create :py:class:`~heyoka.expression` objects
representing :ref:`runtime parameters <runtime_param>`. The parameter index
must be passed to the index operator of the factory object.

Examples:
  >>> from heyoka import par
  >>> p0 = par[0] # p0 will represent the parameter value at index 0

"""


# Machinery for the time attribute.
def _create_time():
    from . import core

    return core._time


time: expression = _create_time()
"""
Time expression.

This global object is an :py:class:`~heyoka.expression` which is used to represent
time (i.e., the independent variable) in differential equations.

"""

eop_data_row: _dtype = eop_data_row
"""
EOP data row.

.. versionadded:: 7.3.0

This is a :ref:`structured NumPy datatype<numpy:defining-structured-types>` used to represent
a row of EOP data in the :py:class:`~heyoka.eop_data` class. The fields in the datatype are:

- the UTC MJD,
- the UT1-UTC difference (in seconds),
- the :math:`x` component of the polar motion (in arcsecs),
- the :math:`y` component of the polar motion (in arcsecs),
- the :math:`x` component of the correction to the IAU 2000/2006
  precession/nutation model (in milliarcsecs),
- the :math:`y` component of the correction to the IAU 2000/2006
  precession/nutation model (in milliarcsecs).

"""

sw_data_row: _dtype = sw_data_row
"""
Space weather data row.

.. versionadded:: 7.3.0

This is a :ref:`structured NumPy datatype<numpy:defining-structured-types>` used to represent
a row of space weather (SW) data in the :py:class:`~heyoka.sw_data` class. The fields in the datatype are:

- the UTC MJD,
- the arithmetic average of the 8 Ap indices for the day,
- the observed 10.7-cm solar radio flux (F10.7),
- the 81-day arithmetic average of the observed F10.7 centred on the day.

"""

del _dtype



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


import unittest as _ut


class elp2000_test_case(_ut.TestCase):
    # Just a small basic test.
    def test_basic(self):
        from . import cfunc
        from .model import elp2000_cartesian_e2000, elp2000_cartesian_fk5

        sol = elp2000_cartesian_e2000(thresh=1e-5)[0]
        cf = cfunc([sol], [])

        date = 2469000.5
        self.assertAlmostEqual(
            cf([], time=(date - 2451545.0) / 36525)[0], -361605.79234692274
        )

        sol = elp2000_cartesian_fk5(thresh=1e-5)[0]
        cf = cfunc([sol], [])

        self.assertAlmostEqual(
            cf([], time=(date - 2451545.0) / 36525)[0], -361605.7668217605
        )



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest as _ut


class model_test_case(_ut.TestCase):
    def test_mascon(self):
        from . import (
            model,
            make_vars,
            expression as ex,
            par,
        )

        x, y, z, vx, vy, vz = make_vars("x", "y", "z", "vx", "vy", "vz")

        dyn = model.mascon(
            Gconst=1.5, masses=[1.1], positions=[[1.0, 2.0, 3.0]], omega=[0.0, 0.0, 3.0]
        )

        self.assertEqual(dyn[0][0], x)
        self.assertEqual(dyn[0][1], ex("vx"))

        pot = model.mascon_potential(
            Gconst=1.5, masses=[1.1], positions=[[1.0, 2.0, 3.0]], omega=[0.0, 0.0, 3.0]
        )

        en = model.mascon_energy(
            Gconst=1.5, masses=[1.1], positions=[[1.0, 2.0, 3.0]], omega=[0.0, 0.0, 3.0]
        )

        self.assertGreater(len(en), len(pot))

        with self.assertRaises(ValueError) as cm:
            model.mascon(
                Gconst=1.5,
                masses=[1.1],
                positions=[1.0, 2.0, 3.0],
                omega=[0.0, 0.0, 3.0],
            )
        self.assertTrue(
            "Invalid positions array in a mascon model: the number of dimensions must"
            " be 2, but it is 1 instead" in str(cm.exception)
        )

        with self.assertRaises(ValueError) as cm:
            model.mascon(
                Gconst=1.5,
                masses=[1.1],
                positions=[[1.0, 2.0, 3.0, 4.0]],
                omega=[0.0, 0.0, 3.0],
            )
        self.assertTrue(
            "Invalid positions array in a mascon model: the number of columns must be"
            " 3, but it is 4 instead" in str(cm.exception)
        )

        with self.assertRaises(TypeError) as cm:
            model.mascon(
                Gconst=1.5,
                masses=[1.1],
                positions=[[{}, {}, {}]],
                omega=[0.0, 0.0, 3.0],
            )
        self.assertTrue(
            "The positions array in a mascon model could not be converted into an array"
            " of expressions - please make sure that the array's values can be"
            " converted into heyoka expressions" in str(cm.exception)
        )

        # Run also a test with parametric mass.
        dyn = model.mascon(
            Gconst=1.5,
            masses=[par[0]],
            positions=[[1.0, 2.0, 3.0]],
            omega=[0.0, 0.0, 3.0],
        )

    def test_rotating(self):
        from . import model, make_vars, expression as ex

        x, y, z, vx, vy, vz = make_vars("x", "y", "z", "vx", "vy", "vz")

        dyn = model.rotating(omega=[0.0, 0.0, 3.0])

        self.assertEqual(dyn[0][0], x)
        self.assertEqual(dyn[0][1], ex("vx"))

        self.assertTrue(len(dyn[3][1]) > 5)

        model.rotating_potential([0.0, 0.0, 3.0])

        model.rotating_energy([0.0, 0.0, 3.0])

    def test_fixed_centres(self):
        from . import model, make_vars, expression as ex
        from numpy import single

        x, y, z, vx, vy, vz = make_vars("x", "y", "z", "vx", "vy", "vz")

        dyn = model.fixed_centres(Gconst=1.5, masses=[1.1], positions=[[1.0, 2.0, 3.0]])

        self.assertEqual(dyn[0][0], x)
        self.assertEqual(dyn[0][1], ex("vx"))

        model.fixed_centres_energy(
            Gconst=1.5, masses=[1.1], positions=[[1.0, 2.0, 3.0]]
        )

        model.fixed_centres_potential(
            Gconst=1.5, masses=[1.1], positions=[[1.0, 2.0, 3.0]]
        )

        with self.assertRaises(ValueError) as cm:
            model.fixed_centres(Gconst=1.5, masses=[1.1], positions=[1.0, 2.0, 3.0])
        self.assertTrue(
            "Invalid positions array in a fixed centres model: the number of dimensions"
            " must be 2, but it is 1 instead" in str(cm.exception)
        )

        with self.assertRaises(ValueError) as cm:
            model.fixed_centres(
                Gconst=1.5, masses=[1.1], positions=[[1.0, 2.0, 3.0, 4.0]]
            )
        self.assertTrue(
            "Invalid positions array in a fixed centres model: the number of columns"
            " must be 3, but it is 4 instead" in str(cm.exception)
        )

        with self.assertRaises(TypeError) as cm:
            model.fixed_centres(Gconst=1.5, masses=[1.1], positions=[[{}, {}, {}]])
        self.assertTrue(
            "The positions array in a fixed centres model could not be converted into"
            " an array of expressions - please make sure that the array's values can be"
            " converted into heyoka expressions" in str(cm.exception)
        )

        # Run also a small test with single-precision values.
        dyn = model.fixed_centres(
            Gconst=single(1.5), masses=[single(1.1)], positions=[[1.0, 2.0, 3.0]]
        )

        self.assertEqual(dyn[0][0], x)
        self.assertEqual(dyn[0][1], ex("vx"))
        self.assertTrue("1.10000002" in repr(dyn))

    def test_nbody(self):
        from . import model, expression, make_vars

        dyn = model.nbody(2, masses=[0.0, 0.0])

        self.assertEqual(len(dyn), 12)

        self.assertEqual(dyn[3][1], expression(0.0))
        self.assertEqual(dyn[4][1], expression(0.0))
        self.assertEqual(dyn[5][1], expression(0.0))

        self.assertEqual(dyn[9][1], expression(0.0))
        self.assertEqual(dyn[10][1], expression(0.0))
        self.assertEqual(dyn[11][1], expression(0.0))

        dyn = model.nbody(2, Gconst=5.0)

        self.assertTrue("5.0000000000000" in str(dyn[3][1]))
        self.assertTrue("5.0000000000000" in str(dyn[4][1]))
        self.assertTrue("5.0000000000000" in str(dyn[5][1]))

        self.assertTrue("5.0000000000000" in str(dyn[9][1]))
        self.assertTrue("5.0000000000000" in str(dyn[10][1]))
        self.assertTrue("5.0000000000000" in str(dyn[11][1]))

        en = model.nbody_energy(2, masses=[0.0, 0.0])

        self.assertEqual(en, expression(0.0))

        en = model.nbody_energy(2, Gconst=5.0)

        self.assertTrue("5.0000000000000" in str(en))

        x0, y0, z0, x1, y1, z1 = make_vars("x_0", "y_0", "z_0", "x_1", "y_1", "z_1")

        dyn = model.np1body(2, masses=[0.0, 0.0])

        self.assertEqual(len(dyn), 6)

        self.assertEqual(dyn[3][1], expression(0.0))
        self.assertEqual(dyn[4][1], expression(0.0))
        self.assertEqual(dyn[5][1], expression(0.0))

        dyn = model.np1body(2, Gconst=5.0)

        self.assertTrue("10.0000000000000" in str(dyn[3][1]))
        self.assertTrue("10.0000000000000" in str(dyn[4][1]))
        self.assertTrue("10.0000000000000" in str(dyn[5][1]))

        en = model.np1body_energy(2, masses=[])

        self.assertEqual(en, expression(0.0))

        en = model.np1body_energy(2, Gconst=5.0)

        self.assertTrue("5.0000000000000" in str(en))

    def test_pendulum(self):
        from . import model, make_vars, sin, cos

        x, v = make_vars("x", "v")

        dyn = model.pendulum()

        self.assertEqual(dyn[0][0], x)
        self.assertEqual(dyn[0][1], v)
        self.assertEqual(dyn[1][0], v)
        self.assertEqual(dyn[1][1], -sin(x))

        dyn = model.pendulum(gconst=2.0)

        self.assertEqual(dyn[0][0], x)
        self.assertEqual(dyn[0][1], v)
        self.assertEqual(dyn[1][0], v)
        self.assertEqual(dyn[1][1], -2.0 * sin(x))

        dyn = model.pendulum(gconst=4.0, length=2.0)

        self.assertEqual(dyn[0][0], x)
        self.assertEqual(dyn[0][1], v)
        self.assertEqual(dyn[1][0], v)
        self.assertEqual(dyn[1][1], -2.0 * sin(x))

        en = model.pendulum_energy()

        self.assertEqual(
            en, ((0.50000000000000000 * v**2) + (1.0000000000000000 - cos(x)))
        )

        en = model.pendulum_energy(gconst=2.0)

        self.assertEqual(
            en,
            (
                (0.50000000000000000 * v**2)
                + (2.0000000000000000 * (1.0000000000000000 - cos(x)))
            ),
        )

        en = model.pendulum_energy(length=2.0, gconst=4.0)

        self.assertEqual(
            en,
            (
                (2.0000000000000000 * v**2)
                + (8.0000000000000000 * (1.0000000000000000 - cos(x)))
            ),
        )

    def test_cr3bp(self):
        from . import model, make_vars

        x, px, y = make_vars("x", "px", "y")

        dyn = model.cr3bp()
        self.assertEqual(dyn[0][0], x)
        self.assertEqual(dyn[0][1], px + y)

        dyn = model.cr3bp(mu=1.0 / 2**4)
        self.assertTrue("0.06250000000" in str(dyn[3][1]))

        jac = model.cr3bp_jacobi()
        self.assertTrue("0.00100000" in str(jac))

        jac = model.cr3bp_jacobi(mu=1.0 / 2**4)
        self.assertTrue("0.06250000000" in str(jac))

    def test_ffnn(self):
        from . import model, make_vars, expression, par, sum as hysum

        x, y = make_vars("x", "y")

        def linear(x):
            return x

        my_ffnn1 = model.ffnn([x], [], 1, [linear])
        my_ffnn2 = model.ffnn([x, y], [], 1, [linear])
        self.assertEqual(my_ffnn1[0], (par[0] * x) + par[1])
        self.assertEqual(my_ffnn2[0], hysum([par[0] * x, par[1] * y, par[2]]))

        my_ffnn3 = model.ffnn([x], [], 1, [linear], [expression(1.2), expression(1.3)])
        self.assertEqual(my_ffnn3[0], expression(1.3) + (expression(1.2) * x))

        my_ffnn4 = model.ffnn([x], [], 1, [linear], [1.2, 1.3])
        self.assertEqual(my_ffnn4[0], expression(1.3) + (expression(1.2) * x))

    def test_cart2geo(self):
        import numpy as np
        from . import model, make_vars, cfunc

        x, y, z = make_vars("x", "y", "z")
        h, phi, lon = make_vars("h", "phi", "lon")
        geodesic1 = model.cart2geo([x, y, z], ecc2=0.13, R_eq=60.0, n_iters=1)
        geodesic2 = model.cart2geo([x, y, z])

        # We test on all inputs (no defaults)
        geodesic1_cf = cfunc(geodesic1, vars=[x, y, z])
        self.assertTrue(
            (geodesic1_cf([1, -1, 1])[0] + 59.791916138446254) ** 2 < 1e-12**2
        )
        self.assertTrue(
            (geodesic1_cf([1, -1, 1])[1] + 0.2053312550471871) ** 2 < 1e-12**2
        )
        self.assertTrue(
            (geodesic1_cf([1, -1, 1])[2] + 0.7853981633974483) ** 2 < 1e-12**2
        )

        # We test that the default values are correctly set
        geodesic2_cf = cfunc(geodesic2, vars=[x, y, z])
        self.assertTrue(
            (geodesic2_cf([6000000, 6000000, 6000000])[0] - 4021307.660867557) ** 2
            < 1e-12**2
        )
        self.assertTrue(
            (geodesic2_cf([6000000, 6000000, 6000000])[1] - 0.6174213396277664) ** 2
            < 1e-12**2
        )
        self.assertTrue(
            (geodesic2_cf([6000000, 6000000, 6000000])[2] - 0.7853981633974483) ** 2
            < 1e-12**2
        )

        # Test the inverse too.
        cart2_cf = cfunc(model.geo2cart([h, phi, lon]), vars=[h, phi, lon])
        geo = geodesic2_cf([6000000, 6000000, 6000000])
        cart = cart2_cf(geo)
        self.assertLess(np.linalg.norm(cart - [6000000, 6000000, 6000000]), 1e-4)

        # Check that the optional args are passed correctly.
        self.assertNotEqual(
            model.geo2cart([h, phi, lon]), model.geo2cart([h, phi, lon], R_eq=1.0)
        )
        self.assertNotEqual(
            model.geo2cart([h, phi, lon]), model.geo2cart([h, phi, lon], ecc2=0.1)
        )
        self.assertNotEqual(
            model.geo2cart([h, phi, lon], R_eq=1.0),
            model.geo2cart([h, phi, lon], ecc2=0.1),
        )

        self.assertNotEqual(
            model.cart2geo([h, phi, lon]), model.cart2geo([h, phi, lon], R_eq=1.0)
        )
        self.assertNotEqual(
            model.cart2geo([h, phi, lon]), model.cart2geo([h, phi, lon], ecc2=0.1)
        )
        self.assertNotEqual(
            model.cart2geo([h, phi, lon], R_eq=1.0),
            model.cart2geo([h, phi, lon], ecc2=0.1),
        )

    def test_nrlmsise00(self):
        from . import model, make_vars, cfunc, time

        h, lat, lon, f107, f107a, ap = make_vars(
            "h", "lat", "lon", "f107", "f107a", "ap"
        )
        nrlmsise00 = model.nrlmsise00_tn(
            geodetic=[h, lat, lon],
            f107=f107,
            f107a=f107a,
            ap=ap,
            time_expr=time / 86400,
        )
        nrlmsise00_cf = cfunc([nrlmsise00], vars=[h, lat, lon, f107, f107a, ap])

        # We test on zero time
        self.assertTrue(
            (
                nrlmsise00_cf([600, 1.2, 3.9, 21.2, 12.2, 22.0], time=0.0)[0]
                - 9.599548606663777e-15
            )
            ** 2
            < 1e-12**2
        )
        # We test some days later
        self.assertTrue(
            (
                nrlmsise00_cf([234, 4.5, 1.02, 4, 3, 5], time=123.23 * 86400.0)[0]
                - 3.549961466488851e-11
            )
            ** 2
            < 1e-12**2
        )

    def test_jb08(self):
        from . import model, make_vars, cfunc, time

        (
            h,
            lat,
            lon,
            f107a,
            f107,
            s107a,
            s107,
            m107a,
            m107,
            y107a,
            y107,
            dDstdT,
        ) = make_vars(
            "h",
            "lat",
            "lon",
            "f107a",
            "f107",
            "s107a",
            "s107",
            "m107a",
            "m107",
            "y107a",
            "y107",
            "dDstdT",
        )
        jb08 = model.jb08_tn(
            geodetic=[h, lat, lon],
            f107=f107,
            f107a=f107a,
            s107=s107,
            s107a=s107a,
            m107=m107,
            m107a=m107a,
            y107=y107,
            y107a=y107a,
            dDstdT=dDstdT,
            time_expr=time / 86400,
        )
        jb08_cf = cfunc(
            [jb08],
            vars=[
                h,
                lat,
                lon,
                f107a,
                f107,
                s107a,
                s107,
                m107a,
                m107,
                y107a,
                y107,
                dDstdT,
            ],
        )

        # We test on zero time
        self.assertTrue(
            (
                jb08_cf([600, 1.2, 3.9, 3, 4, 5, 6, 7, 8, 9, 10, 11], time=0.0)[0]
                - 6.805408788157112e-15
            )
            ** 2
            < 1e-12**2
        )
        # We test some days later
        self.assertTrue(
            (
                jb08_cf(
                    [234, 4.5, 1.02, 11, 10, 9, 8, 7, 6, 5, 4, 3], time=123.23 * 86400.0
                )[0]
                - 1.3364825974582714e-11
            )
            ** 2
            < 1e-12**2
        )

    def test_sgp4(self):
        from . import par, time as tm
        from .model import sgp4

        self.assertEqual(len(sgp4()), 7)
        self.assertEqual(len(sgp4([])), 7)

        # Test also with custom inputs.
        self.assertEqual(len(sgp4(["a", "b", "c", "d", "e", "f", par[0], tm])), 7)

    def test_gpe_is_deep_space(self):
        try:
            from sgp4.api import Satrec
        except ImportError:
            return

        from .model import gpe_is_deep_space

        # A non-deepspace TLE.
        s1 = "1 00045U 60007A   24187.45810325  .00000504  00000-0  14841-3 0  9992"
        t1 = "2 00045  66.6943  81.3521 0257384 317.3173  40.8180 14.34783636277898"
        sat = Satrec.twoline2rv(s1, t1)

        self.assertFalse(gpe_is_deep_space(sat.no_kozai, sat.ecco, sat.inclo))

        # A deepspace TLE.
        t1 = "2 00045  66.6943  81.3521 0257384 317.3173  40.8180  6.34783636277898"
        sat = Satrec.twoline2rv(s1, t1)

        self.assertTrue(gpe_is_deep_space(sat.no_kozai, sat.ecco, sat.inclo))

    def test_era_erap(self):
        from . import cfunc, make_vars
        from .model import era, erap
        import numpy as np

        x = make_vars("x")
        cf = cfunc([era(x), erap(x)], [x])

        out = cf(inputs=[0.0])
        self.assertFalse(np.any(np.isnan(out)))

    def test_pm(self):
        from . import cfunc, make_vars
        from .model import pm_x, pm_xp, pm_y, pm_yp
        import numpy as np

        x = make_vars("x")
        cf = cfunc([pm_x(x), pm_xp(x), pm_y(x), pm_yp(x)], [x])

        out = cf(inputs=[0.0])
        self.assertFalse(np.any(np.isnan(out)))

    def test_sw(self):
        from . import cfunc, make_vars
        from .model import Ap_avg, f107, f107a_center81
        import numpy as np

        x = make_vars("x")
        cf = cfunc([Ap_avg(x), f107(x), f107a_center81(x)], [x])

        out = cf(inputs=[0.0])
        self.assertFalse(np.any(np.isnan(out)))

    def test_dXdY(self):
        from . import cfunc, make_vars
        from .model import dX, dXp, dY, dYp
        import numpy as np

        x = make_vars("x")
        cf = cfunc([dX(x), dXp(x), dY(x), dYp(x)], [x])

        out = cf(inputs=[0.0])
        self.assertFalse(np.any(np.isnan(out)))

    def test_rot_fk5j2000_icrs(self):
        from . import cfunc, make_vars
        from .model import rot_fk5j2000_icrs, rot_icrs_fk5j2000
        import numpy as np

        x, y, z = make_vars("x", "y", "z")
        cf = cfunc(rot_fk5j2000_icrs(rot_icrs_fk5j2000([x, y, z])), [x, y, z])

        out = cf(inputs=[1.0, 1.1, 1.2])
        self.assertTrue(np.allclose(out, [1.0, 1.1, 1.2], rtol=1e-15, atol=0.0))

    def test_time_conversions(self):
        from . import cfunc, make_vars
        from .model import delta_tdb_tt, delta_tt_tai
        import numpy as np

        x = make_vars("x")
        cf = cfunc([delta_tdb_tt(x), delta_tt_tai], [x])

        out = cf(inputs=[0.0])
        self.assertFalse(np.isnan(out[0]))
        self.assertTrue(np.allclose(out[1], 32.184, rtol=1e-16, atol=0.0))

    def test_iau2006(self):
        from . import cfunc, make_vars
        from .model import iau2006
        import numpy as np

        x = make_vars("x")
        cf = cfunc(iau2006(time_expr=x, thresh=1e-6), [x])

        out = cf(inputs=[0.0])
        self.assertFalse(np.isnan(out[0]))

    def test_rot_itrs_icrs(self):
        from . import cfunc, make_vars
        from .model import rot_itrs_icrs, rot_icrs_itrs
        import numpy as np

        x, y, z = make_vars("x", "y", "z")
        cf = cfunc(rot_icrs_itrs(rot_itrs_icrs([x, y, z])), [x, y, z])

        out = cf(inputs=[1.0, 1.1, 1.2], time=0.0)
        self.assertTrue(np.allclose(out, [1.0, 1.1, 1.2], rtol=1e-15, atol=0.0))

        out = cf(inputs=[1.0, 1.1, 1.2], time=0.1)
        self.assertTrue(np.allclose(out, [1.0, 1.1, 1.2], rtol=1e-15, atol=0.0))

        out = cf(inputs=[1.0, 1.1, 1.2], time=-0.1)
        self.assertTrue(np.allclose(out, [1.0, 1.1, 1.2], rtol=1e-15, atol=0.0))

    def test_egm2008(self):
        from . import make_vars
        from .model import egm2008_pot, egm2008_acc, get_egm2008_mu, get_egm2008_a

        x, y, z = make_vars("x", "y", "z")

        self.assertNotEqual(get_egm2008_mu(), get_egm2008_a())
        self.assertNotEqual(
            egm2008_pot([x, y, z], n=2, m=2), egm2008_pot([x, y, z], n=2, m=2, mu=1.2)
        )
        self.assertNotEqual(
            egm2008_pot([x, y, z], n=2, m=2), egm2008_pot([x, y, z], n=2, m=2, a=1.2)
        )
        self.assertNotEqual(
            egm2008_pot([x, y, z], n=2, m=2, a=1.2),
            egm2008_pot([x, y, z], n=2, m=2, mu=1.2),
        )

        self.assertNotEqual(
            egm2008_acc([x, y, z], n=2, m=2), egm2008_acc([x, y, z], n=2, m=2, mu=1.2)
        )
        self.assertNotEqual(
            egm2008_acc([x, y, z], n=2, m=2), egm2008_acc([x, y, z], n=2, m=2, a=1.2)
        )
        self.assertNotEqual(
            egm2008_acc([x, y, z], n=2, m=2, a=1.2),
            egm2008_acc([x, y, z], n=2, m=2, mu=1.2),
        )

    def test_dayfrac(self):
        from . import make_vars
        from .model import dayfrac

        self.assertEqual(str(dayfrac()), "dayfrac(t)")

        x = make_vars("x")
        self.assertEqual(str(dayfrac(time_expr=x)), "dayfrac(x)")

    def test_gmst82(self):
        from . import make_vars
        from .model import gmst82, gmst82p

        self.assertTrue(str(gmst82()).startswith("eop_gmst82_"))
        self.assertTrue(str(gmst82()).endswith("(t)"))

        x = make_vars("x")
        self.assertTrue(str(gmst82(time_expr=x)).endswith("(x)"))

        self.assertTrue(str(gmst82p()).startswith("eop_gmst82p_"))
        self.assertTrue(str(gmst82p(time_expr=x)).endswith("(x)"))

    def test_rot_itrs_teme(self):
        import numpy as np
        from . import cfunc, make_vars
        from .model import rot_itrs_teme, rot_teme_itrs

        x, y, z = make_vars("x", "y", "z")
        teme_x, teme_y, teme_z = rot_itrs_teme([x, y, z])
        cf = cfunc([teme_x, teme_y, teme_z], [x, y, z], compact_mode=True)

        itrs_x, itrs_y, itrs_z = rot_teme_itrs([x, y, z])
        cf_inv = cfunc([itrs_x, itrs_y, itrs_z], [x, y, z], compact_mode=True)

        out = cf(
            [-5821.35967267, 2079.53567927, -2820.30734583], time=0.19938024382589065
        )

        self.assertLess(
            np.linalg.norm(out - [-6102.44327643, -986.33201609, -2820.31307072]), 1e-5
        )

        out = cf_inv(out, time=0.19938024382589065)

        self.assertLess(
            np.linalg.norm(out - [-5821.35967267, 2079.53567927, -2820.30734583]), 1e-10
        )

    def test_rsw(self):
        # A couple of simple tests from orekit.
        from .model import (
            state_to_rsw,
            state_to_rsw_inertial,
            state_from_rsw,
            state_from_rsw_inertial,
        )
        from . import cfunc, make_vars
        import numpy as np

        pos_x, pos_y, pos_z = make_vars("pos_x", "pos_y", "pos_z")
        vel_x, vel_y, vel_z = make_vars("vel_x", "vel_y", "vel_z")
        x, y, z = make_vars("x", "y", "z")
        vx, vy, vz = make_vars("vx", "vy", "vz")

        # state_to_rsw().
        pos_p, vel_p = state_to_rsw(
            pos=[pos_x, pos_y, pos_z],
            vel=[vel_x, vel_y, vel_z],
            r=[x, y, z],
            v=[vx, vy, vz],
        )
        cf_to_rsw = cfunc(
            pos_p + vel_p,
            [pos_x, pos_y, pos_z, vel_x, vel_y, vel_z, x, y, z, vx, vy, vz],
        )

        input = [
            1.9901598679710650e06,
            -4.3171886347018217e05,
            6.7720686840924025e06,
            -6.9420492298495501e03,
            -2.1395918103707527e03,
            1.9120431913109790e03,
            1.9990152502378467e06,
            -4.2466313738494366e05,
            6.7714722017919971e06,
            -6.9397802817958964e03,
            -2.1318724003511638e03,
            1.9205549571233923e03,
        ]

        output = cf_to_rsw(input)

        state_orekit = np.array(
            [
                -1508.0546801836172,
                10340.099985439594,
                4400.563621644753,
                2.6516116955953297,
                3.7183675363633046,
                7.9607554111462075,
            ]
        )

        self.assertTrue(np.all(np.abs((output - state_orekit) / state_orekit) < 1e-12))

        # state_to_rsw_inertial().
        pos_p, vel_p = state_to_rsw_inertial(
            pos=[pos_x, pos_y, pos_z],
            vel=[vel_x, vel_y, vel_z],
            r=[x, y, z],
            v=[vx, vy, vz],
        )
        cf_to_rsw = cfunc(
            pos_p + vel_p,
            [pos_x, pos_y, pos_z, vel_x, vel_y, vel_z, x, y, z, vx, vy, vz],
        )

        input = [
            1.9901598679710650e06,
            -4.3171886347018217e05,
            6.7720686840924025e06,
            -6.9420492298495501e03,
            -2.1395918103707527e03,
            1.9120431913109790e03,
            1.9990152502378467e06,
            -4.2466313738494366e05,
            6.7714722017919971e06,
            -6.9397802817958964e03,
            -2.1318724003511638e03,
            1.9205549571233923e03,
        ]

        output = cf_to_rsw(input)

        state_orekit = np.array(
            [
                -1508.0546801836172,
                10340.099985439594,
                4400.563621644753,
                -3.0126937116365298,
                7511.706686243073,
                7.96075541114692,
            ]
        )

        self.assertTrue(np.all(np.abs((output - state_orekit) / state_orekit) < 1e-12))

        # state_from_rsw().
        pos_p, vel_p = state_from_rsw(
            pos=[pos_x, pos_y, pos_z],
            vel=[vel_x, vel_y, vel_z],
            r=[x, y, z],
            v=[vx, vy, vz],
        )
        cf_to_rsw = cfunc(
            pos_p + vel_p,
            [pos_x, pos_y, pos_z, vel_x, vel_y, vel_z, x, y, z, vx, vy, vz],
        )

        input = [
            -1508.0546801836172,
            10340.099985439594,
            4400.563621644753,
            2.6516116955953297,
            3.7183675363633046,
            7.9607554111462075,
            1.9990152502378467e06,
            -4.2466313738494366e05,
            6.7714722017919971e06,
            -6.9397802817958964e03,
            -2.1318724003511638e03,
            1.9205549571233923e03,
        ]

        output = cf_to_rsw(input)

        state_orekit = np.array(
            [
                1.9901598679710650e06,
                -4.3171886347018217e05,
                6.7720686840924025e06,
                -6.9420492298495501e03,
                -2.1395918103707527e03,
                1.9120431913109790e03,
            ]
        )

        self.assertTrue(np.all(np.abs((output - state_orekit) / state_orekit) < 1e-12))

        # state_from_rsw_inertial().
        pos_p, vel_p = state_from_rsw_inertial(
            pos=[pos_x, pos_y, pos_z],
            vel=[vel_x, vel_y, vel_z],
            r=[x, y, z],
            v=[vx, vy, vz],
        )
        cf_to_rsw = cfunc(
            pos_p + vel_p,
            [pos_x, pos_y, pos_z, vel_x, vel_y, vel_z, x, y, z, vx, vy, vz],
        )

        input = [
            -1508.0546801836172,
            10340.099985439594,
            4400.563621644753,
            -3.0126937116365298,
            7511.706686243073,
            7.96075541114692,
            1.9990152502378467e06,
            -4.2466313738494366e05,
            6.7714722017919971e06,
            -6.9397802817958964e03,
            -2.1318724003511638e03,
            1.9205549571233923e03,
        ]

        output = cf_to_rsw(input)

        state_orekit = np.array(
            [
                1.9901598679710650e06,
                -4.3171886347018217e05,
                6.7720686840924025e06,
                -6.9420492298495501e03,
                -2.1395918103707527e03,
                1.9120431913109790e03,
            ]
        )

        self.assertTrue(np.all(np.abs((output - state_orekit) / state_orekit) < 1e-12))



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


import unittest as _ut


class var_ode_sys_test_case(_ut.TestCase):
    def test_enum(self):
        from . import var_args

        self.assertEqual(var_args.vars | var_args.time | var_args.params, var_args.all)
        self.assertTrue((var_args.vars | var_args.time) & var_args.time)

    def test_basic(self):
        from . import make_vars, var_ode_sys, var_args, sin, par, time
        from copy import copy, deepcopy
        from pickle import dumps, loads

        x, v = make_vars("x", "v")

        orig_sys = [(x, v), (v, -par[0] * sin(x) + time)]

        vsys = var_ode_sys(orig_sys, var_args.vars)

        self.assertEqual(orig_sys, vsys.sys[:2])
        self.assertEqual(vsys.vargs, [x, v])
        self.assertEqual(vsys.n_orig_sv, 2)
        self.assertEqual(vsys.order, 1)

        vsys = var_ode_sys(sys=orig_sys, args=[v, time, x], order=2)

        self.assertEqual(orig_sys, vsys.sys[:2])
        self.assertEqual(vsys.vargs, [v, time, x])
        self.assertEqual(vsys.n_orig_sv, 2)
        self.assertEqual(vsys.order, 2)

        vsys = var_ode_sys(orig_sys, var_args.vars | var_args.params)

        self.assertEqual(orig_sys, vsys.sys[:2])
        self.assertEqual(vsys.vargs, [x, v, par[0]])
        self.assertEqual(vsys.n_orig_sv, 2)
        self.assertEqual(vsys.order, 1)

        vsys2 = copy(vsys)

        self.assertEqual(orig_sys, vsys2.sys[:2])
        self.assertEqual(vsys2.vargs, [x, v, par[0]])
        self.assertEqual(vsys2.n_orig_sv, 2)
        self.assertEqual(vsys2.order, 1)

        vsys2 = deepcopy(vsys)

        self.assertEqual(orig_sys, vsys2.sys[:2])
        self.assertEqual(vsys2.vargs, [x, v, par[0]])
        self.assertEqual(vsys2.n_orig_sv, 2)
        self.assertEqual(vsys2.order, 1)

        vsys2 = loads(dumps(vsys))

        self.assertEqual(orig_sys, vsys2.sys[:2])
        self.assertEqual(vsys2.vargs, [x, v, par[0]])
        self.assertEqual(vsys2.n_orig_sv, 2)
        self.assertEqual(vsys2.order, 1)



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest as _ut


class real128_test_case(_ut.TestCase):
    def test_scalar(self):
        from . import core

        if not hasattr(core, "real128"):
            return

        import random
        from . import real128
        from .core import _ppc_arch

        if _ppc_arch:
            ld = float
        else:
            from numpy import longdouble as ld

        from numpy import float32 as f32

        # Make random testing deterministic.
        random.seed(42)

        # Constructors.
        self.assertEqual(str(real128()), "0")

        # Small ints.
        self.assertEqual(str(real128(1)), "1")
        self.assertEqual(str(real128(-1)), "-1")
        self.assertEqual(str(real128(42)), "42")
        self.assertEqual(str(real128(-42)), "-42")

        # Bool.
        self.assertEqual(real128(True), 1)
        self.assertEqual(real128(False), 0)

        # Large ints, still exactly representable.
        for _ in range(100):
            n = random.randint(-(2**101), 2**101)
            self.assertEqual(str(real128(n)), str(n))

        # Ints with bit width around the mantissa bit width.
        self.assertEqual(str(real128(2**113)), str(2**113))
        self.assertEqual(str(real128(-(2**113))), str(-(2**113)))
        self.assertEqual(str(real128(2**113 + 1)), str(2**113))
        self.assertEqual(str(real128(-(2**113) - 1)), str(-(2**113)))
        self.assertEqual(str(real128(2**113 - 1)), str(2**113 - 1))
        self.assertEqual(str(real128(-(2**113) + 1)), str(-(2**113) + 1))

        # Very large ints.
        for _ in range(100):
            n = random.randint(-(2**1001), 2**1001)
            self.assertEqual(str(real128(n)), str(real128(str(n))))

        # Construction from floats.
        self.assertEqual(str(real128(-0.0)), "-0")
        self.assertEqual(str(real128(42.0)), "42")
        self.assertEqual(str(real128(-float("inf"))), "-inf")
        self.assertEqual(str(real128(float("nan"))), "nan")

        # Construction from long double.
        self.assertEqual(str(real128(ld(-0.0))), "-0")
        self.assertEqual(str(real128(ld(42.0))), "42")
        self.assertEqual(str(real128(ld(-float("inf")))), "-inf")
        self.assertEqual(str(real128(ld(float("nan")))), "nan")

        # Construction from f32.
        self.assertEqual(str(real128(f32(-0.0))), "-0")
        self.assertEqual(str(real128(f32(42.0))), "42")
        self.assertEqual(str(real128(f32(-float("inf")))), "-inf")
        self.assertEqual(str(real128(f32(float("nan")))), "nan")
        self.assertEqual(str(real128(f32("1.1"))), "1.10000002384185791015625")

        # Construction from string.
        self.assertEqual(str(real128("-0")), "-0")
        self.assertEqual(str(real128("-inf")), "-inf")
        self.assertEqual(str(real128("42")), "42")
        self.assertEqual(str(real128("-123")), "-123")

        # Construction from real.
        if hasattr(core, "real"):
            real = core.real

            self.assertEqual(real128(real("1.1", 113)), real128("1.1"))
            self.assertNotEqual(real128(real("1.1", 100)), real128("1.1"))

        with self.assertRaises(ValueError) as cm:
            real128("hello world")
        self.assertTrue(
            "The string 'hello world' does not represent a valid quadruple-precision"
            " floating-point value" in str(cm.exception)
        )

        with self.assertRaises(TypeError) as cm:
            real128([])
        self.assertTrue(
            'Cannot construct a real128 from an object of type "list"'
            in str(cm.exception)
        )

        # Check that the constructor takes exactly 1 argument.
        with self.assertRaises(TypeError) as cm:
            real128(1, prec=7)
        self.assertTrue("function takes at most 1 argument" in str(cm.exception))

        with self.assertRaises(TypeError) as cm:
            real128(1, 2)
        self.assertTrue("function takes at most 1 argument" in str(cm.exception))

        with self.assertRaises(TypeError) as cm:
            real128(prec=7)

        # Conversion to bool.
        self.assertTrue(bool(real128(1)))
        self.assertTrue(bool(real128(-1)))
        self.assertFalse(bool(real128(0)))
        self.assertTrue(bool(real128("inf")))
        self.assertTrue(bool(real128("nan")))
        self.assertTrue(bool(real128("-inf")))

        # Unary math operations.
        self.assertEqual(repr(+real128(1)), "1")
        self.assertEqual(repr(+real128(-1)), "-1")
        self.assertEqual(repr(-real128(42)), "-42")
        self.assertEqual(repr(-real128(-42)), "42")
        self.assertEqual(repr(abs(real128(-42))), "42")

        # Binary math ops.
        self.assertEqual(repr(real128(42) + real128(1)), "43")
        self.assertEqual(repr(real128(42) + 1), "43")
        self.assertEqual(repr(real128(42) + 1.0), "43")
        # self.assertEqual(repr(real128(42) + ld(1)), "43")
        self.assertEqual(repr(real128(42) + f32(1)), "43")
        self.assertEqual(repr(1 + real128(42)), "43")
        self.assertEqual(repr(1.0 + real128(42)), "43")
        # self.assertEqual(repr(ld(1) + real128(42)), "43")
        self.assertEqual(repr(f32(1) + real128(42)), "43")
        if hasattr(core, "real"):
            real = core.real
            self.assertTrue(isinstance(real128() + real(), real))
            self.assertTrue(isinstance(real() + real128(), real))
            self.assertEqual(
                real128("1.1") + real("1.1", 100), real("1.1", 113) + real("1.1", 100)
            )
            self.assertEqual(
                real("1.1", 100) + real128("1.1"), real("1.1", 113) + real("1.1", 100)
            )
        with self.assertRaises(TypeError) as cm:
            real128(1) + []
        with self.assertRaises(TypeError) as cm:
            [] + real128(1)

        # A small test to check that the numpy operators
        # do not interfere with our own.
        self.assertEqual(str(real128(0) + f32("1.1")), "1.10000002384185791015625")
        self.assertEqual(str(f32("1.1") + real128(0)), "1.10000002384185791015625")

        self.assertEqual(repr(real128(42) - real128(1)), "41")
        self.assertEqual(repr(real128(42) - 1), "41")
        self.assertEqual(repr(real128(42) - 1.0), "41")
        # self.assertEqual(repr(real128(42) - ld(1)), "41")
        self.assertEqual(repr(real128(42) - f32(1)), "41")
        self.assertEqual(repr(1 - real128(42)), "-41")
        self.assertEqual(repr(1.0 - real128(42)), "-41")
        # self.assertEqual(repr(ld(1) - real128(42)), "-41")
        self.assertEqual(repr(f32(1) - real128(42)), "-41")
        if hasattr(core, "real"):
            real = core.real
            self.assertTrue(isinstance(real128() - real(), real))
            self.assertTrue(isinstance(real() - real128(), real))
            self.assertEqual(
                real128("1.1") - real("1.1", 100), real("1.1", 113) - real("1.1", 100)
            )
            self.assertEqual(
                real("1.1", 100) - real128("1.1"), real("1.1", 100) - real("1.1", 113)
            )
        with self.assertRaises(TypeError) as cm:
            real128(1) - []
        with self.assertRaises(TypeError) as cm:
            [] - real128(1)

        self.assertEqual(repr(real128(42) * real128(2)), "84")
        self.assertEqual(repr(real128(42) * 2), "84")
        self.assertEqual(repr(real128(42) * 2.0), "84")
        # self.assertEqual(repr(real128(42) * ld(2)), "84")
        self.assertEqual(repr(real128(42) * f32(2)), "84")
        self.assertEqual(repr(2 * real128(42)), "84")
        self.assertEqual(repr(2.0 * real128(42)), "84")
        # self.assertEqual(repr(ld(2) * real128(42)), "84")
        self.assertEqual(repr(f32(2) * real128(42)), "84")
        if hasattr(core, "real"):
            real = core.real
            self.assertTrue(isinstance(real128() * real(), real))
            self.assertTrue(isinstance(real() * real128(), real))
            self.assertEqual(
                real128("1.1") * real("1.1", 100), real("1.1", 113) * real("1.1", 100)
            )
            self.assertEqual(
                real("1.1", 100) * real128("1.1"), real("1.1", 113) * real("1.1", 100)
            )
        with self.assertRaises(TypeError) as cm:
            real128(1) * []
        with self.assertRaises(TypeError) as cm:
            [] * real128(1)

        self.assertEqual(repr(real128(42) // real128(9)), "4")
        self.assertEqual(repr(real128(42) // 9), "4")
        self.assertEqual(repr(real128(42) // 9.0), "4")
        # self.assertEqual(repr(real128(42) // ld(9)), "4")
        self.assertEqual(repr(real128(42) // f32(9)), "4")
        self.assertEqual(repr(-42 // real128(9)), "-5")
        self.assertEqual(repr(-42.0 // real128(9)), "-5")
        # self.assertEqual(repr(ld(-42) // real128(9)), "-5")
        self.assertEqual(repr(f32(-42) // real128(9)), "-5")
        if hasattr(core, "real"):
            real = core.real
            self.assertTrue(isinstance(real128(1) // real(1), real))
            self.assertTrue(isinstance(real(1) // real128(1), real))
            self.assertEqual(
                real128("1.1") // real("1.1", 100), real("1.1", 113) // real("1.1", 100)
            )
            self.assertEqual(
                real("1.1", 100) // real128("1.1"), real("1.1", 100) // real("1.1", 113)
            )
        with self.assertRaises(TypeError) as cm:
            real128(1) // []
        with self.assertRaises(TypeError) as cm:
            [] // real128(1)

        self.assertEqual(repr(real128(42) ** real128(2)), "1764")
        self.assertEqual(repr(real128(42) ** 2), "1764")
        self.assertEqual(repr(real128(42) ** 2.0), "1764")
        # self.assertEqual(repr(real128(42) ** ld(2)), "1764")
        self.assertEqual(repr(real128(42) ** f32(2)), "1764")
        self.assertEqual(repr(42 ** real128(2)), "1764")
        self.assertEqual(repr(42.0 ** real128(2)), "1764")
        # self.assertEqual(repr(ld(42) ** real128(2)), "1764")
        self.assertEqual(repr(f32(42) ** real128(2)), "1764")
        with self.assertRaises(TypeError) as cm:
            real128(1) ** []
        with self.assertRaises(TypeError) as cm:
            [] ** real128(1)
        with self.assertRaises(ValueError) as cm:
            pow(real128(1), 3, mod=4)
        self.assertTrue(
            "Modular exponentiation is not supported for real128" in str(cm.exception)
        )

        # Comparisons.
        self.assertTrue(real128(1) < real128(2))
        self.assertTrue(real128(1) < 2)
        self.assertTrue(real128(1) < 2.0)
        # self.assertTrue(real128(1) < ld(2))
        self.assertTrue(real128(1) < f32(2))
        self.assertTrue(1 < real128(2))
        self.assertTrue(1.0 < real128(2))
        # self.assertTrue(ld(1) < real128(2))
        self.assertTrue(f32(1) < real128(2))
        self.assertFalse(real128("nan") < 2)
        self.assertFalse(2 < real128("nan"))
        self.assertFalse(real128("nan") < real128("nan"))
        if hasattr(core, "real"):
            real = core.real

            self.assertTrue(real(1) < real128(2))
            self.assertTrue(real128(1) < real(2))
            self.assertFalse(real(1) > real128(2))
            self.assertFalse(real128(1) > real(2))
        with self.assertRaises(TypeError) as cm:
            real128(1) < []
        with self.assertRaises(TypeError) as cm:
            [] < real128(1)

        # The codepath for the other comparisons is identical,
        # let's limit to some light testing.
        self.assertTrue(real128(1) <= real128(2))
        self.assertTrue(real128(2) <= real128(2))
        self.assertFalse(real128(3) <= real128(2))
        self.assertTrue(real128(2) == real128(2))
        self.assertFalse(real128(3) == real128(2))
        self.assertFalse(real128(3) == real128("NaN"))
        self.assertFalse(real128(2) != real128(2))
        self.assertTrue(real128(3) != real128(2))
        self.assertTrue(real128(3) != real128("NaN"))
        self.assertTrue(real128(3) > real128(2))
        self.assertFalse(real128(2) > real128(3))
        self.assertFalse(real128(1) >= real128(2))
        self.assertTrue(real128(2) >= real128(2))
        self.assertTrue(real128(3) >= real128(2))

        # Copy/deepcopy.
        from copy import copy, deepcopy

        x = real128("1.1")
        y = copy(x)
        self.assertEqual(x, y)
        y = deepcopy(x)
        self.assertEqual(x, y)

        # Pickling.
        # NOTE: this is provided for free by the
        # NumPy inheritance.
        import pickle

        x = real128("1.1")
        y = pickle.loads(pickle.dumps(x))
        self.assertEqual(x, y)

    def test_numpy(self):
        from . import core

        if not hasattr(core, "real128"):
            return

        import numpy as np
        from copy import copy, deepcopy
        from . import real128
        from pickle import dumps, loads

        # Basic creation/getitem/setitem.
        arr = np.array([1, 2, 3], dtype=real128)
        self.assertEqual(type(arr[0]), real128)
        self.assertEqual(arr[0], 1)
        self.assertEqual(arr[1], 2)
        self.assertEqual(arr[2], 3)

        arr = np.array([real128("1.1")])
        self.assertEqual(arr.dtype, real128)
        self.assertEqual(type(arr[0]), real128)
        self.assertEqual(arr[0], real128("1.1"))
        arr[0] = real128("1.3")
        self.assertEqual(arr[0], real128("1.3"))
        arr[0] = real128(123)
        self.assertEqual(arr[0], 123)
        arr[0] = real128(1.1)
        self.assertEqual(arr[0], real128(1.1))

        with self.assertRaises(TypeError) as cm:
            arr = np.array(["1.1"], dtype=np.dtype(real128))
        self.assertTrue(
            "Cannot invoke __setitem__() on a real128 array with an input value of type"
            ' "str"' in str(cm.exception)
        )

        # Copying primitives.
        arr = np.array([real128("1.1"), real128("1.3")])
        arr_copy = copy(arr)
        self.assertTrue(arr_copy[0] == real128("1.1"))
        self.assertTrue(arr_copy[1] == real128("1.3"))
        arr_copy = deepcopy(arr)
        self.assertTrue(arr_copy[0] == real128("1.1"))
        self.assertTrue(arr_copy[1] == real128("1.3"))

        # Nonzero.
        arr = np.array([1, 0, 3, real128("nan")], dtype=real128)
        self.assertTrue(np.all([0, 2, 3] == np.nonzero(arr)[0]))

        # Fill.
        arr = np.zeros((10,), dtype=real128)
        arr.fill(real128("1.1"))
        for _ in arr:
            self.assertEqual(_, real128("1.1"))

        # Argmin/argmax.
        arr = np.array([1, 321, 54, 6, 2, 6, -6], dtype=real128)
        self.assertEqual(np.argmin(arr), 6)
        self.assertEqual(np.argmax(arr), 1)
        arr = np.array([], dtype=real128)
        with self.assertRaises(ValueError) as cm:
            np.argmin(arr)
        with self.assertRaises(ValueError) as cm:
            np.argmax(arr)

        # arange() and linspace().
        arr = np.arange(0, 1, real128("0.3"))
        self.assertTrue(
            np.all(
                arr
                == np.array(
                    [
                        0,
                        real128("0.29999999999999999999999999999999999"),
                        real128("0.599999999999999999999999999999999981"),
                        real128("0.899999999999999999999999999999999923"),
                    ],
                    dtype=real128,
                )
            )
        )
        self.assertEqual(arr.dtype, real128)
        arr = np.linspace(real128(0), 1, 4, dtype=real128)
        self.assertTrue(
            np.all(
                arr
                == np.array(
                    [
                        0,
                        real128("0.333333333333333333333333333333333317"),
                        real128("0.666666666666666666666666666666666635"),
                        1,
                    ],
                    dtype=real128,
                )
            )
        )
        self.assertEqual(arr.dtype, real128)

        # zeros, ones, full.
        arr = np.zeros((2, 2), dtype=real128)
        self.assertTrue(np.all(arr == np.array([[0, 0], [0, 0]], dtype=real128)))
        arr = np.ones((2, 2), dtype=real128)
        self.assertTrue(np.all(arr == np.array([[1, 1], [1, 1]], dtype=real128)))
        arr = np.full((2, 2), real128("1.1"))
        self.assertTrue(
            np.all(
                arr
                == np.array(
                    [[real128("1.1"), real128("1.1")], [real128("1.1"), real128("1.1")]]
                )
            )
        )

        # dot product.
        arr1 = np.array([real128("1.1"), real128("1.3")])
        arr2 = np.array([real128("2.1"), real128("2.3")])
        self.assertEqual(
            real128("1.1") * real128("2.1") + real128("1.3") * real128("2.3"),
            np.dot(arr1, arr2),
        )
        arr1 = np.array([], dtype=real128)
        arr2 = np.array([], dtype=real128)
        self.assertEqual(0, np.dot(arr1, arr2))

        # Matrix multiplication.
        mat = np.array(
            [[real128("1.1"), real128("1.3")], [real128("2.1"), real128("2.3")]]
        )
        self.assertTrue(
            np.all(
                mat @ mat
                == np.array(
                    [
                        [
                            real128("3.94000000000000000000000000000000034"),
                            real128("4.41999999999999999999999999999999994"),
                        ],
                        [
                            real128("7.14000000000000000000000000000000049"),
                            real128("8.01999999999999999999999999999999963"),
                        ],
                    ]
                )
            )
        )

        # Conversions.
        arr = np.array([real128("1.1")])
        self.assertEqual(arr.astype(float)[0], 1.1)
        self.assertEqual(arr.astype(np.int32)[0], 1)
        self.assertEqual(arr.astype(bool)[0], True)
        arr = np.array([1], dtype=np.int32)
        self.assertEqual(arr.astype(real128)[0], real128(1))
        self.assertEqual(arr.astype(real128, casting="safe")[0], real128(1))
        arr = np.array([1.1], dtype=float)
        self.assertEqual(arr.astype(real128)[0], real128(1.1))
        self.assertEqual(arr.astype(real128, casting="safe")[0], real128(1.1))
        arr = np.array([real128("1.1")])
        with self.assertRaises(TypeError) as cm:
            arr.astype(float, casting="safe")
        with self.assertRaises(TypeError) as cm:
            arr.astype(np.int32, casting="safe")

        # Arithmetic.
        a = real128("1.1")
        b = real128("1.3")
        c = real128("-.7")
        intval = 2
        arr1 = np.array([a])
        arr2 = np.array([b])
        arr3 = np.array([c])
        arrint = np.array([2], dtype=np.int64)

        self.assertEqual((arr1 + arr2)[0], a + b)
        self.assertEqual((arr1 + arrint)[0], a + intval)
        self.assertEqual((arrint + arr2)[0], intval + b)

        self.assertEqual((arr1 - arr2)[0], a - b)
        self.assertEqual((arr1 - arrint)[0], a - intval)
        self.assertEqual((arrint - arr2)[0], intval - b)

        self.assertEqual((arr1 * arr2)[0], a * b)
        self.assertEqual((arr1 * arrint)[0], a * intval)
        self.assertEqual((arrint * arr2)[0], intval * b)

        self.assertEqual((arr1 / arr2)[0], a / b)
        self.assertEqual((arr1 / arrint)[0], a / intval)
        self.assertEqual((arrint / arr2)[0], intval / b)

        self.assertEqual(np.square(arr1)[0], a * a)

        self.assertEqual((arr1 // arr2)[0], a // b)
        self.assertEqual((arr1 // arrint)[0], a // intval)
        self.assertEqual((arrint // arr2)[0], intval // b)

        self.assertEqual((+arr1)[0], a)
        self.assertEqual((-arr1)[0], -a)

        self.assertEqual(abs(-arr1)[0], a)
        self.assertEqual(np.abs(-arr1)[0], a)
        self.assertEqual(np.fabs(-arr1)[0], a)

        # Power/roots.
        self.assertEqual((arr1**arr2)[0], a**b)
        self.assertEqual((arr1**arrint)[0], a**intval)
        self.assertEqual((arrint**arr2)[0], intval**b)

        self.assertEqual(
            np.sqrt(arr1)[0], real128("1.04880884817015154699145351367993759")
        )

        self.assertEqual(
            np.cbrt(arr1)[0], real128("1.03228011545636715921358522500970152")
        )

        # Trigonometry.
        self.assertEqual(
            np.sin(arr1)[0], real128("0.891207360061435339951802577871703605")
        )

        self.assertEqual(
            np.cos(arr1)[0], real128("0.453596121425577387771370051784716053")
        )

        self.assertEqual(
            np.tan(arr1)[0], real128("1.96475965724865195093092278177937863")
        )

        self.assertEqual(
            np.arcsin(arr3)[0], real128("-0.775397496610753063740353352714987079")
        )

        self.assertEqual(
            np.arccos(arr3)[0], real128("2.34619382340564968297167504435473857")
        )

        self.assertEqual(
            np.arctan(arr3)[0], real128("-0.610725964389208616543758876490236037")
        )

        self.assertEqual(
            np.arctan2(arr1, arr2)[0], real128("0.702256931509007079704992531169094469")
        )

        self.assertEqual(
            np.arctan2(arr1, arrint)[0],
            real128("0.502843210927860827330882029245277632"),
        )
        self.assertEqual(
            np.arctan2(arrint, arr2)[0],
            real128("0.994421106203712939003751213245981381"),
        )

        self.assertEqual(
            np.sinh(arr1)[0], real128("1.33564747012417677938478052357867843")
        )

        self.assertEqual(
            np.cosh(arr1)[0], real128("1.6685185538222563326736274300099942")
        )

        self.assertEqual(
            np.tanh(arr1)[0], real128("0.800499021760629706011461330600696557")
        )

        self.assertEqual(
            np.arcsinh(arr3)[0], real128("-0.652666566082355786808686344109675833")
        )

        self.assertEqual(
            np.arccosh(arr1)[0], real128("0.443568254385115189132911066352498299")
        )

        self.assertEqual(
            np.arctanh(arr3)[0], real128("-0.867300527694053194427144690475300431")
        )

        self.assertEqual(
            np.deg2rad(np.array([1], dtype=real128))[0],
            real128(
                "0.01745329251994329576923690768488612713442871888541725456097191440171005"
            ),
        )

        self.assertEqual(
            np.radians(np.array([1], dtype=real128))[0],
            real128(
                "0.01745329251994329576923690768488612713442871888541725456097191440171005"
            ),
        )

        self.assertEqual(
            np.rad2deg(np.array([1], dtype=real128))[0],
            real128(
                "57.295779513082320876798154814105170332405472466564321549160243861202985"
            ),
        )

        self.assertEqual(
            np.degrees(np.array([1], dtype=real128))[0],
            real128(
                "57.295779513082320876798154814105170332405472466564321549160243861202985"
            ),
        )

        # Exponentials and logarithms.
        self.assertEqual(
            np.exp(arr1)[0], real128("3.00416602394643311205840795358867282")
        )

        self.assertEqual(
            np.expm1(arr1)[0], real128("2.00416602394643311205840795358867243")
        )

        self.assertEqual(
            np.log(arr1)[0], real128("0.0953101798043248600439521232807651556")
        )

        self.assertEqual(
            np.log2(arr1)[0], real128("0.137503523749934908329043617236402896")
        )

        self.assertEqual(
            np.log10(arr1)[0], real128("0.0413926851582250407501999712430242757")
        )

        self.assertEqual(
            np.log1p(arr1)[0], real128("0.741937344729377312482606525681341284")
        )

        # Comparisons.
        self.assertEqual((arr1 < arr2).dtype, bool)
        self.assertTrue((arr1 < arr2)[0])
        self.assertEqual((arr1 < arrint).dtype, bool)
        self.assertTrue((arr1 < arrint)[0])

        self.assertEqual((arr1 <= arr1).dtype, bool)
        self.assertTrue((arr1 <= arr2)[0])
        self.assertEqual((arr1 <= arrint).dtype, bool)
        self.assertTrue((arr1 <= arrint)[0])

        self.assertEqual((arr1 == arr1).dtype, bool)
        self.assertFalse((arr1 == arr2)[0])
        self.assertEqual((arr1 == arrint).dtype, bool)
        self.assertFalse((arr1 == arrint)[0])

        self.assertEqual((arr1 != arr1).dtype, bool)
        self.assertTrue((arr1 != arr2)[0])
        self.assertEqual((arr1 != arrint).dtype, bool)
        self.assertTrue((arr1 != arrint)[0])

        self.assertEqual((arr1 > arr2).dtype, bool)
        self.assertTrue((arr2 > arr1)[0])
        self.assertEqual((arrint > arr1).dtype, bool)
        self.assertTrue((arrint > arr1)[0])

        self.assertEqual((arr1 >= arr1).dtype, bool)
        self.assertTrue((arr2 >= arr1)[0])
        self.assertEqual((arrint >= arr1).dtype, bool)
        self.assertTrue((arrint >= arr1)[0])

        # isfinite().
        self.assertEqual((np.isfinite(arr1)).dtype, bool)
        self.assertTrue((np.isfinite(arr1))[0])
        self.assertTrue((np.isfinite([real128("1"), real128("nan")]))[0])
        self.assertFalse((np.isfinite([real128("1"), real128("nan")]))[1])
        self.assertFalse((np.isfinite([real128("1"), real128("+inf")]))[1])

        # isinf().
        self.assertEqual((np.isinf(arr1)).dtype, bool)
        self.assertFalse((np.isinf(arr1))[0])
        self.assertFalse((np.isinf([real128("1"), real128("nan")]))[0])
        self.assertTrue((np.isinf([real128("1"), real128("-inf")]))[1])
        self.assertTrue((np.isinf([real128("1"), real128("+inf")]))[1])

        # sign.
        self.assertEqual(np.sign(-real128("1.1")), -1)
        self.assertEqual(np.sign(real128("0")), 0)
        self.assertEqual(np.sign(real128("1.1")), 1)
        self.assertEqual((np.sign(arr1)).dtype, real128)
        self.assertTrue(
            np.all(np.sign([real128("10"), real128("0"), real128("-10")]) == [1, 0, -1])
        )

        # maximum/minimum.
        self.assertTrue(
            np.all(
                np.minimum([10, real128("1.1"), 20], [1, real128("1.11"), 25])
                == [1, real128("1.1"), 20]
            )
        )
        self.assertTrue(
            np.all(
                np.maximum([10, real128("1.1"), 20], [1, real128("1.11"), 25])
                == [10, real128("1.11"), 25]
            )
        )

        # Pickling.
        tmp = np.array([real128("1.1"), real128("1.11"), real128("1.13")])
        self.assertTrue(np.all(tmp == loads(dumps(tmp))))

        # Use hstack to check the copyswap implementation.
        arr1 = np.array([real128("1.1"), real128("1.11"), real128("1.13")])
        arr2 = np.array([real128("2.1"), real128("2.11"), real128("2.13")])
        arr3 = np.hstack([arr1, arr2])

        self.assertTrue(np.all(arr3[:3] == arr1))
        self.assertTrue(np.all(arr3[3:] == arr2))

        # Sorting (via the compare primitive).
        arr1 = np.array([real128("3"), real128("1"), real128("2")])
        arr1_sorted = np.sort(arr1)
        self.assertTrue(
            np.all(arr1_sorted == [real128("1"), real128("2"), real128("3")])
        )

        # Check NaN handling.
        arr1 = np.array([real128("nan"), real128("1"), real128("nan")])
        self.assertTrue(all(np.isnan(arr1) == [True, False, True]))
        arr1_sorted = np.sort(arr1)
        self.assertEqual(arr1_sorted[0], 1)
        self.assertFalse(np.isnan(arr1_sorted[0]))
        self.assertTrue(np.isnan(arr1_sorted[1]))
        self.assertTrue(np.isnan(arr1_sorted[2]))

        # Setitem from real.
        if hasattr(core, "real"):
            real = core.real

            arr1 = np.array([1, 2, 3], dtype=real128)
            arr1[1] = real("1.1", 113)
            self.assertEqual(arr1[1], real128("1.1"))



if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Test suite runner.")
    parser.add_argument("--with-numba", action="store_true")

    args = parser.parse_args()

    import heyoka

    if args.with_numba:
        import numba

    heyoka.test.run_test_suite()

    if hasattr(heyoka, "real"):
        heyoka.install_custom_numpy_mem_handler()
        heyoka.test.run_test_suite()
        heyoka.remove_custom_numpy_mem_handler()
        heyoka.test.run_test_suite()



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

_with_sympy = True

try:
    import sympy as _spy
except ImportError:
    _with_sympy = False


def _from_sympy_symbol(sym):
    import re

    # Check if it is a parameter.
    m = re.match(r"par\[((?:[1-9][0-9]*|0))\]", sym.name)

    if m:
        from . import par

        return par[int(m.groups()[0])]
    else:
        from . import expression

        return expression(sym.name)


def _from_sympy_number(ex):
    is_rational = isinstance(ex, _spy.Rational)

    if (
        not isinstance(ex, _spy.Float)
        and not isinstance(ex, _spy.Integer)
        and not is_rational
    ):
        raise TypeError(
            "Only floating-point, integer and (some) rational numbers can be converted"
            " from sympy"
        )

    from . import expression

    # Extract the needed precision in bits.
    # NOTE: the bit size returned by mpmath accounts
    # for the implicit bit and it is thus consistent
    # with the value returned by bit_length().
    if is_rational:
        # NOTE: for rationals we allow conversion only
        # if den is a power of 2.
        den = ex.q
        if not (den & (den - 1) == 0):
            raise ValueError(
                "Cannot convert from sympy a rational number whose denominator is not a"
                " power of 2"
            )

        # The needed precision is given by the bit size of the
        # numerator.
        prec = ex.p.bit_length()
    else:
        prec = (
            ex.num.context.prec if isinstance(ex, _spy.Float) else int(ex).bit_length()
        )

    nf_err_msg = "A non-finite value was produced when converting from a sympy number"

    if prec <= 53:
        # Double precision is sufficient to represent
        # exactly the number.

        from math import isfinite

        retval = float(ex)

        # NOTE: a non-finite value could be produced if the original
        # number is non-finite or if its exponent is too large.
        if not isfinite(retval):
            raise ValueError(nf_err_msg)

        return expression(retval)

    import numpy as np

    # NOTE: the number returned by finfo does not account for
    # the implicit bit.
    if prec <= np.finfo(np.longdouble).nmant + 1:
        # Long double precision is sufficient to represent
        # exactly the number.
        retval = (
            np.longdouble(ex.p) / np.longdouble(ex.q)
            if is_rational
            else np.longdouble(str(ex))
        )
        if not np.isfinite(retval):
            raise ValueError(nf_err_msg)

        return expression(retval)

    from . import core

    if hasattr(core, "real128") and prec <= 113:
        # We have real128, and quadmath precision
        # is enough to represent exactly the number.
        real128 = core.real128

        retval = real128(ex.p) / real128(ex.q) if is_rational else real128(str(ex))

        if not np.isfinite(retval):
            raise ValueError(nf_err_msg)

        return expression(retval)

    if hasattr(core, "real"):
        # We have real, we can in principle represent
        # any number.
        real = core.real

        # Ensure we are not going to employ
        # a too-low precision.
        prec = max(prec, core.real_prec_min())

        retval = (
            real(ex.p, prec) / real(ex.q, prec) if is_rational else real(str(ex), prec)
        )

        if not np.isfinite(retval):
            raise ValueError(nf_err_msg)

        return expression(retval)

    raise ValueError(
        "Cannot convert the number {} from sympy exactly: the required precision ({})"
        " is too high".format(ex, prec)
    )


def _build_fmap():
    if not _with_sympy:
        return None

    from . import core, pi, time as htime

    retval = {}

    retval[_spy.acos] = core.acos
    retval[_spy.acosh] = core.acosh
    retval[_spy.asin] = core.asin
    retval[_spy.asinh] = core.asinh
    retval[_spy.atan] = core.atan
    retval[_spy.atan2] = core.atan2
    retval[_spy.atanh] = core.atanh
    retval[_spy.cos] = core.cos
    retval[_spy.cosh] = core.cosh
    retval[_spy.erf] = core.erf
    retval[_spy.exp] = core.exp
    retval[_spy.log] = core.log
    retval[_spy.sin] = core.sin
    retval[_spy.sinh] = core.sinh
    retval[_spy.tan] = core.tan
    retval[_spy.tanh] = core.tanh
    retval[_spy.Pow] = lambda x, y: x**y
    # NOTE: sympy.pi is an instance of this type.
    retval[_spy.core.numbers.Pi] = lambda: pi

    def add_wrapper(*args):
        return core.sum(args)

    retval[_spy.Add] = add_wrapper

    def mul_wrapper(*args):
        return core.prod(args)

    retval[_spy.Mul] = mul_wrapper

    retval[_spy.Function("heyoka_kepE")] = core.kepE
    retval[_spy.Function("heyoka_kepF")] = core.kepF
    retval[_spy.Function("heyoka_kepDE")] = core.kepDE
    retval[_spy.Function("heyoka_time")] = lambda: htime

    return retval


_fmap = _build_fmap()


def _from_sympy_function(func, s_dict, c_dict):
    args = [_from_sympy_impl(arg, s_dict, c_dict) for arg in func.args]

    tp = type(func)

    if not tp in _fmap:
        raise TypeError("Unable to convert the sympy object {}".format(func))

    return _fmap[tp](*args)


def _from_sympy_impl(ex, s_dict, c_dict):
    # Check s_dict first.
    if ex in s_dict:
        return s_dict[ex]

    # Check if we already converted this expression.
    if id(ex) in c_dict:
        return c_dict[id(ex)]

    if isinstance(ex, _spy.Number):
        ret = _from_sympy_number(ex)
        c_dict[id(ex)] = ret
        return ret

    if isinstance(ex, _spy.Symbol):
        ret = _from_sympy_symbol(ex)
        c_dict[id(ex)] = ret
        return ret

    ret = _from_sympy_function(ex, s_dict, c_dict)
    c_dict[id(ex)] = ret

    return ret



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


# NOTE: this is a small helper to splat a 1D grid (i.e., a grid for
# a scalar integrator) into the appropriate shape for a batch integrator.
# If ta is a scalar integrator, the original grid will be returned unchanged.
def _splat_grid(arg, ta):
    if hasattr(ta, "batch_size"):
        import numpy as np

        return np.repeat(arg, ta.batch_size).reshape((-1, ta.batch_size))
    else:
        return arg


# Thread-based implementation.
def _ensemble_propagate_thread(tp, ta, arg, n_iter, gen, **kwargs):
    from concurrent.futures import ThreadPoolExecutor
    from copy import deepcopy, copy

    # Pop the multithreading options from kwargs.
    max_workers = kwargs.pop("max_workers", None)

    # Make deep copies of the callback argument, if present.
    if "callback" in kwargs:
        kwargs_list = []

        for i in range(n_iter):
            # Make a shallow copy of the original kwargs.
            # new_kwargs will be a new dict containing
            # references to the objects stored in kwargs.
            new_kwargs = copy(kwargs)

            # Update the callback argument in new_kwargs
            # with a deep copy of the original callback object in
            # kwargs.
            new_kwargs.update(callback=deepcopy(kwargs["callback"]))

            kwargs_list.append(new_kwargs)
    else:
        kwargs_list = [kwargs] * n_iter

    # The worker function.
    def func(i):
        # Create the local integrator.
        local_ta = gen(deepcopy(ta), i)

        # Run the propagation.
        if tp == "until":
            loc_ret = local_ta.propagate_until(arg, **kwargs_list[i])
        elif tp == "for":
            loc_ret = local_ta.propagate_for(arg, **kwargs_list[i])
        else:
            loc_ret = local_ta.propagate_grid(_splat_grid(arg, ta), **kwargs_list[i])

        # Return the results.
        return (local_ta,) + loc_ret

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        ret = list(executor.map(func, range(n_iter)))

    return ret


# The worker function used in the multiprocessing implementation.
def _mp_propagate(tup):
    from . import _s11n_backend_map

    tp, ta, gen, arg, kwargs, i, s11n_str = tup

    # Fetch the s11n backend from its
    # str representation.
    s11n_be = _s11n_backend_map[s11n_str]

    # Unpickle the other arguments.
    ta = s11n_be.loads(ta)
    gen = s11n_be.loads(gen)
    kwargs = s11n_be.loads(kwargs)

    # Create the local integrator.
    local_ta = gen(ta, i)

    # Run the propagation.
    if tp == "until":
        loc_ret = local_ta.propagate_until(arg, **kwargs)
    elif tp == "for":
        loc_ret = local_ta.propagate_for(arg, **kwargs)
    else:
        loc_ret = local_ta.propagate_grid(_splat_grid(arg, ta), **kwargs)

    # Return the results.
    return s11n_be.dumps((local_ta,) + loc_ret)


# Process-based implementation.
def _ensemble_propagate_process(tp, ta, arg, n_iter, gen, **kwargs):
    from concurrent.futures import ProcessPoolExecutor
    import multiprocessing as mp
    from . import get_serialization_backend, _s11n_backend_inv_map

    # Fetch the currently active s11n backend.
    s11n_be = get_serialization_backend()

    # Fetch its string counterpart.
    s11n_str = _s11n_backend_inv_map[s11n_be]

    # NOTE: ensure the processes are started with
    # the 'spawn' method.
    ctx = mp.get_context("spawn")

    # Pop the multiprocessing options from kwargs.
    max_workers = kwargs.pop("max_workers", None)
    chunksize = kwargs.pop("chunksize", 1)

    with ProcessPoolExecutor(max_workers=max_workers, mp_context=ctx) as executor:
        ret = list(
            executor.map(
                _mp_propagate,
                zip(
                    [tp] * n_iter,
                    [s11n_be.dumps(ta)] * n_iter,
                    [s11n_be.dumps(gen)] * n_iter,
                    [arg] * n_iter,
                    [s11n_be.dumps(kwargs)] * n_iter,
                    range(n_iter),
                    [s11n_str] * n_iter,
                ),
                chunksize=chunksize,
            )
        )

    return [s11n_be.loads(_) for _ in ret]



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


import unittest as _ut


class kepE_test_case(_ut.TestCase):
    def test_expr(self):
        from . import kepE, diff, make_vars, sin, cos, core
        from .core import _ppc_arch
        import numpy as np

        x, y = make_vars("x", "y")

        # Try a few overloads.
        kepE(x, y)
        kepE(0.1, y)
        kepE(e=x, M=0.2)

        self.assertEqual(
            diff(kepE(x, y), x), sin(kepE(x, y)) / (1.0 - x * cos(kepE(x, y)))
        )
        self.assertEqual(diff(kepE(x, y), y), 1.0 / (1.0 - x * cos(kepE(x, y))))

        # noconvert() behaviour.
        with self.assertRaises(TypeError) as cm:
            kepE(23, x)

        with self.assertRaises(TypeError) as cm:
            kepE(x, 23)

        if not _ppc_arch:
            self.assertEqual(
                diff(kepE(x, np.longdouble("1.1")), x),
                sin(kepE(x, np.longdouble("1.1")))
                / (1.0 - x * cos(kepE(x, np.longdouble("1.1")))),
            )
            self.assertEqual(
                diff(kepE(np.longdouble("1.1"), y), y),
                1.0 / (1.0 - np.longdouble("1.1") * cos(kepE(np.longdouble("1.1"), y))),
            )

            kepE(np.longdouble(0.1), y)
            kepE(x, np.longdouble(0.2))

            with self.assertRaises(TypeError) as cm:
                kepE(0.1, np.longdouble("1.1"))
            self.assertTrue(
                "At least one of the arguments of kepE() must be an expression"
                in str(cm.exception)
            )

        with self.assertRaises(TypeError) as cm:
            kepE(0.1, 0.2)
        self.assertTrue(
            "At least one of the arguments of kepE() must be an expression"
            in str(cm.exception)
        )

        if not hasattr(core, "real128"):
            return

        from .core import real128

        self.assertEqual(
            diff(kepE(x, real128("1.1")), x),
            sin(kepE(x, real128("1.1"))) / (1.0 - x * cos(kepE(x, real128("1.1")))),
        )
        self.assertEqual(
            diff(kepE(real128("1.1"), y), y),
            1.0 / (1.0 - real128("1.1") * cos(kepE(real128("1.1"), y))),
        )


class kepF_test_case(_ut.TestCase):
    def test_expr(self):
        from . import kepF, make_vars, core
        from .core import _ppc_arch
        import numpy as np

        x, y, z = make_vars("x", "y", "z")

        # Try a few overloads.
        kepF(x, y, z)
        kepF(0.1, y, z)
        kepF(h=0.1, k=0.2, lam=z)

        # noconvert() behaviour.
        with self.assertRaises(TypeError) as cm:
            kepF(23, x, y)

        with self.assertRaises(TypeError) as cm:
            kepF(x, 23, y)

        with self.assertRaises(TypeError) as cm:
            kepF(x, y, 23)

        if not _ppc_arch:
            kepF(x, y, np.longdouble("1.1"))
            kepF(x, np.longdouble(".1"), np.longdouble("1.1"))

            with self.assertRaises(TypeError) as cm:
                kepF(x, 0.1, np.longdouble("1.1"))
            self.assertTrue(
                "The numerical arguments of kepF() must be all of the same type"
                in str(cm.exception)
            )

        with self.assertRaises(TypeError) as cm:
            kepF(0.1, 0.2, 0.3)
        self.assertTrue(
            "At least one of the arguments of kepF() must be an expression"
            in str(cm.exception)
        )

        if not hasattr(core, "real128"):
            return

        from .core import real128

        kepF(real128(0.1), y, z)
        kepF(real128(0.1), real128(0.2), z)

        with self.assertRaises(TypeError) as cm:
            kepF(x, 0.1, real128("1.1"))
        self.assertTrue(
            "The numerical arguments of kepF() must be all of the same type"
            in str(cm.exception)
        )


class kepDE_test_case(_ut.TestCase):
    def test_expr(self):
        from . import kepDE, make_vars, core
        from .core import _ppc_arch
        import numpy as np

        x, y, z = make_vars("x", "y", "z")

        # Try a few overloads.
        kepDE(x, y, z)
        kepDE(0.1, y, z)
        kepDE(s0=0.1, c0=0.2, DM=z)

        # noconvert() behaviour.
        with self.assertRaises(TypeError) as cm:
            kepDE(23, x, y)

        with self.assertRaises(TypeError) as cm:
            kepDE(x, 23, y)

        with self.assertRaises(TypeError) as cm:
            kepDE(x, y, 23)

        if not _ppc_arch:
            kepDE(x, y, np.longdouble("1.1"))
            kepDE(x, np.longdouble(".1"), np.longdouble("1.1"))

            with self.assertRaises(TypeError) as cm:
                kepDE(x, 0.1, np.longdouble("1.1"))
            self.assertTrue(
                "The numerical arguments of kepDE() must be all of the same type"
                in str(cm.exception)
            )

        with self.assertRaises(TypeError) as cm:
            kepDE(0.1, 0.2, 0.3)
        self.assertTrue(
            "At least one of the arguments of kepDE() must be an expression"
            in str(cm.exception)
        )

        if not hasattr(core, "real128"):
            return

        from .core import real128

        kepDE(real128(0.1), y, z)
        kepDE(real128(0.1), real128(0.2), z)

        with self.assertRaises(TypeError) as cm:
            kepDE(x, 0.1, real128("1.1"))
        self.assertTrue(
            "The numerical arguments of kepDE() must be all of the same type"
            in str(cm.exception)
        )



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


import unittest as _ut


class sw_data_test_case(_ut.TestCase):
    def test_basic(self):
        from . import sw_data, sw_data_row
        import numpy as np
        from sys import getrefcount
        import pickle
        from copy import copy, deepcopy

        self.assertTrue(isinstance(sw_data_row, np.dtype))

        # Check access to the data table.
        data = sw_data()
        rc = getrefcount(data)
        tbl = data.table
        self.assertEqual(getrefcount(data), rc + 1)

        # Check that we cannot write to the table.
        with self.assertRaises(ValueError) as cm:
            tbl[:] = tbl[:]
        self.assertTrue("read-only" in str(cm.exception))

        self.assertEqual(tbl.dtype, sw_data_row)

        self.assertGreater(len(data.timestamp), 0)
        self.assertGreater(len(data.identifier), 0)
        self.assertGreater(len(repr(data)), 0)

        # Pickling.
        new_data = pickle.loads(pickle.dumps(data))
        self.assertEqual(new_data.timestamp, data.timestamp)
        self.assertEqual(new_data.identifier, data.identifier)
        self.assertTrue(np.all(new_data.table == tbl))

        # Copy/deepcopy.
        new_data = copy(data)
        self.assertEqual(new_data.timestamp, data.timestamp)
        self.assertEqual(new_data.identifier, data.identifier)
        self.assertTrue(np.all(new_data.table == tbl))

        new_data = deepcopy(data)
        self.assertEqual(new_data.timestamp, data.timestamp)
        self.assertEqual(new_data.identifier, data.identifier)
        self.assertTrue(np.all(new_data.table == tbl))



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest as _ut


class sgp4_propagator_test_case(_ut.TestCase):
    # Couple of TLEs downloaded from SpaceTrack.
    s1 = "1 00045U 60007A   24187.45810325  .00000504  00000-0  14841-3 0  9992"
    t1 = "2 00045  66.6943  81.3521 0257384 317.3173  40.8180 14.34783636277898"

    s2 = "1 00046U 60007B   24187.58585517  .00002292  00000-0  46268-3 0  9993"
    t2 = "2 00046  66.6912   4.9995 0194928 108.0242 254.2215 14.52078523353568"

    def test_basics(self):
        try:
            from sgp4.api import Satrec
        except ImportError:
            return

        from .model import sgp4_propagator
        from pickle import loads, dumps
        from copy import copy, deepcopy
        from . import make_vars, code_model
        import numpy as np

        s1 = sgp4_propagator_test_case.s1
        t1 = sgp4_propagator_test_case.t1

        s2 = sgp4_propagator_test_case.s2
        t2 = sgp4_propagator_test_case.t2

        sat1 = Satrec.twoline2rv(s1, t1)
        sat2 = Satrec.twoline2rv(s2, t2)

        with self.assertRaises(TypeError) as cm:
            sgp4_propagator([sat1, 1])
        self.assertTrue(
            "Invalid object encountered in the satellite data for an sgp4 propagator: a"
            " list of sgp4 Satrec objects is expected, but an object of type '<class"
            " 'int'>' was detected instead at index 1" in str(cm.exception)
        )

        prop = sgp4_propagator([sat1, sat2])

        self.assertEqual(prop.nsats, 2)
        self.assertEqual(prop.nouts, 7)
        self.assertEqual(prop.diff_order, 0)

        with self.assertRaises(ValueError) as cm:
            prop.diff_args
        self.assertTrue(
            "The function 'get_diff_args()' cannot be invoked on an sgp4 propagator"
            " without derivatives" in str(cm.exception)
        )

        self.assertEqual(prop.sat_data.shape, (9, 2))
        self.assertEqual(prop.sat_data[6, 0], 0.14841e-3)
        self.assertEqual(prop.sat_data[6, 1], 0.46268e-3)

        # Check that we cannot write into sat_data.
        with self.assertRaises(ValueError) as cm:
            prop.sat_data[0, 0] = 0

        # Check that we cannot fetch diff properties if derivatives
        # are not requested.
        with self.assertRaises(ValueError) as cm:
            prop.get_dslice(order=1)
        self.assertTrue(
            "The function 'get_dslice()' cannot be invoked on an sgp4 propagator"
            " without derivatives" in str(cm.exception)
        )

        with self.assertRaises(ValueError) as cm:
            prop.get_mindex(i=0)
        self.assertTrue(
            "The function 'get_mindex()' cannot be invoked on an sgp4 propagator"
            " without derivatives" in str(cm.exception)
        )

        self.assertTrue("SGP4 propagator" in repr(prop))
        self.assertTrue("N of satellites: 2" in repr(prop))

        prop.foobar = [1, 2, 3]
        cprop = copy(prop)
        self.assertEqual(cprop.foobar, [1, 2, 3])
        self.assertEqual(id(cprop.foobar), id(prop.foobar))
        self.assertEqual(cprop.sat_data.shape, (9, 2))
        self.assertEqual(cprop.sat_data[6, 0], 0.14841e-3)
        self.assertEqual(cprop.sat_data[6, 1], 0.46268e-3)
        cprop = deepcopy(prop)
        self.assertEqual(cprop.foobar, [1, 2, 3])
        self.assertNotEqual(id(cprop.foobar), id(prop.foobar))
        self.assertEqual(cprop.sat_data.shape, (9, 2))
        self.assertEqual(cprop.sat_data[6, 0], 0.14841e-3)
        self.assertEqual(cprop.sat_data[6, 1], 0.46268e-3)

        cprop = loads(dumps(prop))
        self.assertEqual(cprop.foobar, [1, 2, 3])
        self.assertNotEqual(id(cprop.foobar), id(prop.foobar))
        self.assertEqual(cprop.sat_data.shape, (9, 2))
        self.assertEqual(cprop.sat_data[6, 0], 0.14841e-3)
        self.assertEqual(cprop.sat_data[6, 1], 0.46268e-3)

        # A couple of tests with the derivatives and custom llvm settings.
        prop = sgp4_propagator(
            [sat1, sat2],
            diff_order=1,
            parjit=False,
            compact_mode=True,
            code_model=code_model.large,
        )
        self.assertEqual(
            prop.diff_args,
            make_vars("n0", "e0", "i0", "node0", "omega0", "m0", "bstar"),
        )
        self.assertEqual(prop.get_dslice(order=1), slice(7, 56, None))
        self.assertEqual(prop.get_mindex(7), [0, 1, 0, 0, 0, 0, 0, 0])

        # Test with GPE data passed in as array.
        with self.assertRaises(ValueError) as cm:
            sgp4_propagator(np.zeros((2, 2), dtype=float))
        self.assertTrue(
            "The array of input GPEs for an sgp4 propagator must have 9 rows, but the"
            " supplied array has 2 row(s) instead" in str(cm.exception)
        )
        with self.assertRaises(ValueError) as cm:
            sgp4_propagator(np.zeros((2,), dtype=float))
        self.assertTrue(
            "The array of input GPEs for an sgp4 propagator must have 2 dimensions, but"
            " the supplied array has 1 dimension(s) instead" in str(cm.exception)
        )
        with self.assertRaises(ValueError) as cm:
            sgp4_propagator(np.zeros((18, 2), dtype=float)[::2])
        self.assertTrue(
            "Invalid array of input GPEs detected in an sgp4 propagator: the array is"
            " not C-style contiguous, please consider using numpy.ascontiguousarray()"
            " to turn it into one" in str(cm.exception)
        )

        # Construct a GPE data array from the satellites sat1 and sat2.
        sat_attrs = [
            "no_kozai",
            "ecco",
            "inclo",
            "nodeo",
            "argpo",
            "mo",
            "bstar",
            "jdsatepoch",
            "jdsatepochF",
        ]
        gpe_data = np.ascontiguousarray(
            np.transpose(
                np.array(
                    [
                        [getattr(sat, name) for name in sat_attrs]
                        for sat in [sat1, sat2]
                    ],
                    dtype=float,
                )
            )
        )

        # Init the propagator and verify its gpe data matches prop.
        prop2 = sgp4_propagator(gpe_data)
        self.assertTrue(np.all(prop.sat_data == prop2.sat_data))

    def test_propagation(self):
        try:
            from sgp4.api import Satrec
        except ImportError:
            return

        import numpy as np
        from .model import sgp4_propagator

        s1 = sgp4_propagator_test_case.s1
        t1 = sgp4_propagator_test_case.t1

        s2 = sgp4_propagator_test_case.s2
        t2 = sgp4_propagator_test_case.t2

        sat1 = Satrec.twoline2rv(s1, t1)
        sat2 = Satrec.twoline2rv(s2, t2)

        for fp_type in [float, np.single]:
            prop = sgp4_propagator([sat1, sat2], diff_order=1, fp_type=fp_type)

            # Error checking.
            with self.assertRaises(ValueError) as cm:
                prop(np.zeros((), dtype=fp_type))
            self.assertTrue(
                "A times/dates array with 1 or 2 dimensions is expected as an input for"
                " the call operator of an sgp4 propagator, but an array with 0"
                " dimensions was provided instead" in str(cm.exception)
            )

            with self.assertRaises(ValueError) as cm:
                prop(np.zeros((0, 0, 0), dtype=fp_type))
            self.assertTrue(
                "A times/dates array with 1 or 2 dimensions is expected as an input for"
                " the call operator of an sgp4 propagator, but an array with 3"
                " dimensions was provided instead" in str(cm.exception)
            )

            with self.assertRaises(ValueError) as cm:
                prop(np.zeros((1, 5), dtype=fp_type))
            self.assertTrue(
                "Invalid times/dates array detected as an input for the call operator"
                " of an sgp4 propagator in batch mode: the number of satellites"
                " inferred from the times/dates array is 5, but the propagator contains"
                " 2 satellite(s) instead" in str(cm.exception)
            )

            with self.assertRaises(ValueError) as cm:
                prop(np.zeros((1,), dtype=fp_type))
            self.assertTrue(
                "Invalid times/dates array detected as an input for the call operator"
                " of an sgp4 propagator: the number of satellites inferred from the"
                " times/dates array is 1, but the propagator contains 2 satellite(s)"
                " instead" in str(cm.exception)
            )

            tmp = np.zeros((10, 2), dtype=fp_type)
            tmp2 = tmp[::2, :]

            with self.assertRaises(ValueError) as cm:
                prop(tmp2)
            self.assertTrue(
                "Invalid times/dates array detected as an input for the call operator"
                " of an sgp4 propagator: the array is not C-style contiguous"
                in str(cm.exception)
            )

            # Invalid output provided.
            with self.assertRaises(ValueError) as cm:
                prop(
                    np.zeros((2,), dtype=fp_type),
                    out=np.zeros((100,), dtype=fp_type)[::2],
                )
            self.assertTrue(
                "Invalid output array detected in the call operator of "
                "an sgp4 propagator: the array is not C-style contiguous and writeable"
                in str(cm.exception)
            )

            out = np.zeros(
                (
                    49,
                    2,
                ),
                dtype=fp_type,
            )
            out.flags.writeable = False
            with self.assertRaises(ValueError) as cm:
                prop(
                    np.zeros((2,), dtype=fp_type),
                    out=out,
                )
            self.assertTrue(
                "Invalid output array detected in the call operator of "
                "an sgp4 propagator: the array is not C-style contiguous and writeable"
                in str(cm.exception)
            )

            # Memory overlap.
            out = np.zeros((2,), dtype=fp_type)
            with self.assertRaises(ValueError) as cm:
                prop(
                    out,
                    out=out,
                )
            self.assertTrue(
                "Invalid input/output arrays detected in the call operator of "
                "an sgp4 propagator: the input/outputs arrays may overlap"
                in str(cm.exception)
            )

            # Scalar mode errors.
            out = np.zeros(
                (49,),
                dtype=fp_type,
            )
            with self.assertRaises(ValueError) as cm:
                prop(
                    np.zeros((2,), dtype=fp_type),
                    out=out,
                )
            self.assertTrue(
                "Invalid output array detected in the call operator of "
                "an sgp4 propagator: the array has 1 dimension(s), "
                "but 2 dimensions are expected instead" in str(cm.exception)
            )

            out = np.zeros(
                (48, 2),
                dtype=fp_type,
            )
            with self.assertRaises(ValueError) as cm:
                prop(
                    np.zeros((2,), dtype=fp_type),
                    out=out,
                )
            self.assertTrue(
                "Invalid output array detected in the call operator of "
                "an sgp4 propagator: the first dimension has a "
                "size of 48, but a "
                "size of 56 (i.e., equal to the number of outputs for each "
                "propagation) is required instead" in str(cm.exception)
            )

            out = np.zeros(
                (56, 1),
                dtype=fp_type,
            )
            with self.assertRaises(ValueError) as cm:
                prop(
                    np.zeros((2,), dtype=fp_type),
                    out=out,
                )
            self.assertTrue(
                "Invalid output array detected in the call operator of "
                "an sgp4 propagator: the second dimension has a "
                "size of 1, but a "
                "size of 2 (i.e., equal to the total number of satellites) is "
                "required instead" in str(cm.exception)
            )

            # Batch mode errors.
            out = np.zeros(
                (56, 1),
                dtype=fp_type,
            )
            with self.assertRaises(ValueError) as cm:
                prop(
                    np.zeros((10, 2), dtype=fp_type),
                    out=out,
                )
            self.assertTrue(
                "Invalid output array detected in the call operator of "
                "an sgp4 propagator in batch mode: the array has 2 dimension(s), "
                "but 3 dimensions are expected instead" in str(cm.exception)
            )

            out = np.zeros(
                (9, 56, 1),
                dtype=fp_type,
            )
            with self.assertRaises(ValueError) as cm:
                prop(
                    np.zeros((10, 2), dtype=fp_type),
                    out=out,
                )
            self.assertTrue(
                "Invalid output array detected in the call operator of an sgp4"
                " propagator in batch mode: the first dimension has a size of 9, but a"
                " size of 10 (i.e., equal to the number of evaluations) is required"
                " instead" in str(cm.exception)
            )

            out = np.zeros(
                (10, 48, 1),
                dtype=fp_type,
            )
            with self.assertRaises(ValueError) as cm:
                prop(
                    np.zeros((10, 2), dtype=fp_type),
                    out=out,
                )
            self.assertTrue(
                "Invalid output array detected in the call operator of "
                "an sgp4 propagator in batch mode: the second dimension has a "
                "size of 48, but a "
                "size of 56 (i.e., equal to the number of outputs for each "
                "propagation) is required instead" in str(cm.exception)
            )

            out = np.zeros(
                (10, 56, 0),
                dtype=fp_type,
            )
            with self.assertRaises(ValueError) as cm:
                prop(
                    np.zeros((10, 2), dtype=fp_type),
                    out=out,
                )
            self.assertTrue(
                "Invalid output array detected in the call operator of "
                "an sgp4 propagator in batch mode: the third dimension has a "
                "size of 0, but a "
                "size of 2 (i.e., equal to the total number of satellites) is "
                "required instead" in str(cm.exception)
            )

            # Run a simple propagation to the TLE epoch, check
            # that all derivatives wrt the error codes are zero.
            out = np.zeros(
                (56, 2),
                dtype=fp_type,
            )
            prop(
                np.zeros((2,), dtype=fp_type),
                out=out,
            )
            self.assertTrue(np.all(out[49:, :] == 0.0))

            # Same with a jdate.
            dates = np.zeros((2,), dtype=prop.jdtype)
            dates["jd"] = 2460496.5
            dates["frac"] = 0.5833449099999939
            prop(
                dates,
                out=out,
            )
            self.assertTrue(np.all(out[49:, :] == 0.0))

            # Test that a batch prop with zero nevals does not throw.
            dates = np.zeros((0, 2), dtype=prop.jdtype)
            out = np.zeros(
                (0, 56, 2),
                dtype=fp_type,
            )
            prop(
                dates,
                out=out,
            )

    def test_skyfield_comp(self):
        try:
            from skyfield.api import load
            from skyfield.iokit import parse_tle_file
        except ImportError:
            return

        from .model import sgp4_propagator
        from ._sgp4_test_data import sgp4_test_tle
        import math, numpy as np
        from sgp4.api import SatrecArray

        # Load the test dataset.
        ts = load.timescale()
        satellites = list(
            parse_tle_file((bytes(_, "ascii") for _ in sgp4_test_tle.split("\n")), ts)
        )

        # Filter out high-altitude satellites
        sats = list(
            filter(
                lambda sat: 2 * math.pi / sat.no_kozai < 225,
                (_.model for _ in satellites),
            )
        )

        # Create the propagator.
        prop = sgp4_propagator(sats)

        # Create the dates array.
        dates = np.zeros((prop.nsats,), dtype=prop.jdtype)

        # Pick a date.
        jd = 2460496.5
        jd_frac = 0.5833449099999939

        # Run the test between one month before and after the
        # reference date.
        for delta_day in range(-30, 30, 2):
            dates["jd"] = jd + delta_day
            dates["frac"] = jd_frac

            # Compute the state vectors.
            sv = prop(dates)

            # Compute them with the sgp4 module.
            sat_arr = SatrecArray(sats)
            e, r, v = sat_arr.sgp4(np.array([jd + delta_day]), np.array([jd_frac]))

            # Mask out the satellites that:
            # - generated an error code, or
            # - ended up farther than 8000km from the Earth, or
            # - contain non-finite positional data.
            mask = np.logical_and.reduce(
                (
                    e[:, 0] == 0,
                    np.linalg.norm(r[:, 0, :], axis=1) < 8000,
                    np.all(np.isfinite(r[:, 0, :]), axis=1),
                )
            )

            # Compute the positional errors in meters, and sort them.
            err = np.sort(np.linalg.norm(r[mask, 0] - sv[:3, mask].T, axis=1) * 1e3)

            # Check that the largest error is less than 1mm.
            self.assertTrue(err[-1] < 1e-3)

            # Check the error codes.
            self.assertTrue(np.all(e[:, 0] == sv[-1, :].T))

    def test_replace_sat_data(self):
        try:
            from sgp4.api import Satrec
        except ImportError:
            return

        from copy import deepcopy
        from .model import sgp4_propagator
        import numpy as np

        s1 = sgp4_propagator_test_case.s1
        t1 = sgp4_propagator_test_case.t1

        s2 = sgp4_propagator_test_case.s2
        t2 = sgp4_propagator_test_case.t2

        sat1 = Satrec.twoline2rv(s1, t1)
        sat2 = Satrec.twoline2rv(s2, t2)

        prop = sgp4_propagator([sat1, sat2])
        orig_prop = deepcopy(prop)

        orig_sat_data = deepcopy(prop.sat_data)
        orig_sat_data_ref = prop.sat_data

        prop.replace_sat_data([sat2, sat1])

        # Check array identity.
        self.assertTrue(np.shares_memory(prop.sat_data, orig_sat_data_ref))

        self.assertTrue(np.all(prop.sat_data[:, 1] == orig_sat_data[:, 0]))
        self.assertTrue(np.all(prop.sat_data[:, 0] == orig_sat_data[:, 1]))

        new_sat_data = deepcopy(prop.sat_data)

        with self.assertRaises(TypeError) as cm:
            prop.replace_sat_data([sat1, 1])
        self.assertTrue(
            "Invalid object encountered in the satellite data for an sgp4 propagator: a"
            " list of sgp4 Satrec objects is expected, but an object of type '<class"
            " 'int'>' was detected instead at index 1" in str(cm.exception)
        )

        with self.assertRaises(ValueError) as cm:
            prop.replace_sat_data([sat1])
        self.assertTrue(
            "Invalid array provided to replace_sat_data(): the number of "
            "columns (1) does not match the number of satellites (2)"
            in str(cm.exception)
        )

        self.assertTrue(np.all(prop.sat_data == new_sat_data))

        # Check also the overload with gpe data as a numpy array.

        # Construct a GPE data array from the satellites sat1 and sat2.
        sat_attrs = [
            "no_kozai",
            "ecco",
            "inclo",
            "nodeo",
            "argpo",
            "mo",
            "bstar",
            "jdsatepoch",
            "jdsatepochF",
        ]
        gpe_data = np.ascontiguousarray(
            np.transpose(
                np.array(
                    [
                        [getattr(sat, name) for name in sat_attrs]
                        for sat in [sat1, sat2]
                    ],
                    dtype=float,
                )
            )
        )

        # Init the propagator and verify its gpe data matches prop.
        prop.replace_sat_data(gpe_data)
        self.assertTrue(np.all(prop.sat_data == orig_prop.sat_data))

    def test_leap_second(self):
        # Test with a TLE close to a leap second day.
        try:
            from sgp4.api import Satrec
        except ImportError:
            return

        from .model import sgp4_propagator
        import numpy as np

        s = "1 00045U 60007A   05363.79166667  .00000504  00000-0  14841-3 0  9992"
        t = "2 00045  66.6943  81.3521 0257384 317.3173  40.8180 14.34783636277898"

        sat = Satrec.twoline2rv(s, t)

        prop = sgp4_propagator([sat])

        # Propagate for 10 Julian days, well past year's end.
        dates = np.zeros((1,), dtype=prop.jdtype)
        dates["jd"] = sat.jdsatepoch + 10
        dates["frac"] = sat.jdsatepochF
        res1 = prop(dates)

        # Propagate with a tsince of 10 days plus 1 second
        # (the leap second).
        res2 = prop(np.array([1440.0 * 10 + 1 / 60.0]))

        # Check closeness.
        self.assertTrue(np.allclose(res1[:3, 0], res2[:3, 0], atol=1e-12))
        self.assertTrue(np.allclose(res1[3:, 0], res2[3:, 0], atol=1e-15))
